"""Read from and write to edgelists."""

from ..generators import empty_hypergraph

__all__ = [
    "read_edgelist",
    "write_edgelist",
    "parse_edgelist",
]


def generate_edgelist(H, delimiter=" "):
    """
    A helper function to generate a hyperedge list from a Hypergraph object.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members

    Yields
    -------
    iterator of strings
        Each entry is a line for the file to write.
    """
    for id in H.edges:
        e = H.edges.members(id)
        yield delimiter.join(map(str, e))


def write_edgelist(H, path, delimiter=" ", encoding="utf-8"):
    """Create a file containing a hyperedge list from a Hypergraph object.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    path: string
        The path of the file to write to
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members
    encoding: string, default: "utf-8"
        Encoding of the file

    Examples
    --------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.01, 0.001])
    >>> # xgi.write_edgelist(H, "test.csv", delimiter=",")

    """
    with open(path, "wb") as file:
        for line in generate_edgelist(H, delimiter):
            line += "\n"
            file.write(line.encode(encoding))


def read_edgelist(
    path,
    comments="#",
    delimiter=None,
    create_using=None,
    nodetype=None,
    encoding="utf-8",
):
    """Read a file containing a hyperedge list and
    convert it to a Hypergraph object.

    Parameters
    ----------
    path: string
        The path of the file to read from
    comments: string, default: "#"
        The token that denotes comments in the file
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    nodetype: type
        type that the node labels will be cast to
    encoding: string, default: "utf-8"
        Encoding of the file

    Returns
    -------
    Hypergraph object
        The loaded hypergraph

    See Also
    --------
    read_weighted_edgelist

    Examples
    --------
    >>> import xgi
    >>> # H = xgi.read_edgelist("test.csv", delimiter=",")

    """
    with open(path, "rb") as file:
        lines = (
            line if isinstance(line, str) else line.decode(encoding) for line in file
        )
        return parse_edgelist(
            lines,
            comments=comments,
            delimiter=delimiter,
            create_using=create_using,
            nodetype=nodetype,
        )


def parse_edgelist(
    lines, comments="#", delimiter=None, create_using=None, nodetype=None
):
    """
    A helper function to read a iterable of strings containing a hyperedge list and
    convert it to a Hypergraph object.

    Parameters
    ----------
    lines: iterable of strings
        Lines where each line is an edge
    comments: string, default: "#"
        The token that denotes comments to ignore
    delimiter: char, default: space (" ")
        Specifies the delimiter between hyperedge members
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    nodetype: type
        type that the node labels will be cast to

    Returns
    -------
    Hypergraph object
        The loaded hypergraph

    """
    H = empty_hypergraph(create_using)
    for line in lines:
        if comments is not None:
            p = line.find(comments)
            if p >= 0:
                line = line[:p]
            if not line:
                continue
        edge = line.strip().split(delimiter)

        if nodetype is not None:
            try:
                edge = [nodetype(node) for node in edge]
            except ValueError as e:
                raise TypeError(f"Failed to convert nodes to type {nodetype}.") from e

        H.add_edge(edge)
    return H



"""Statistics of networks, their nodes, and edges.

Any mapping that assigns some quantity to each node of a network is considered a node
statistic.  For example, the degree is a node-integer mapping, while a node attribute
that assigns a string label to each node is a node-string mapping.  The `stats` package
provides a common interface to all such mappings.

Each such mapping is accessible via the `H.nodes` view.  For example, the degree of all
nodes supports type conversion using the `as*` methods.

>>> import xgi
>>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
>>> H.nodes.degree.asdict()
{1: 1, 2: 2, 3: 3, 4: 2, 5: 2}
>>> H.nodes.degree.aslist()
[1, 2, 3, 2, 2]

Another feature is the ability to filter the nodes of a network by degree.

>>> H.nodes.filterby('degree', 2)
NodeView((2, 4, 5))

The power of the stats package is that any other node statistic that can be conceived of
as a node-quantity mapping is given the same interface.  For example, node attributes
get the same treatment:

>>> H.add_nodes_from([
...     (1, {"color": "red", "name": "horse"}),
...     (2, {"color": "blue", "name": "pony"}),
...     (3, {"color": "yellow", "name": "zebra"}),
...     (4, {"color": "red", "name": "orangutan", "age": 20}),
...     (5, {"color": "blue", "name": "fish", "age": 2}),
... ])
>>> H.nodes.attrs('color').asdict()
{1: 'red', 2: 'blue', 3: 'yellow', 4: 'red', 5: 'blue'}
>>> H.nodes.attrs('color').aslist()
['red', 'blue', 'yellow', 'red', 'blue']
>>> H.nodes.filterby_attr('color', 'red')
NodeView((1, 4))

Many other features are available, including edge-statistics, and user-defined
statistics.  For more details, see the `tutorial
<https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

"""

import numpy as np
import pandas as pd
from scipy.stats import moment as spmoment

from ..exception import IDNotFound
from ..utils import hist

from . import edgestats, diedgestats, dinodestats, nodestats

__all__ = [
    "nodestat_func",
    "edgestat_func",
    "dinodestat_func",
    "diedgestat_func",
    "dispatch_stat",
    "dispatch_many_stats",
]


class IDStat:
    """Mapping between nodes or edges and a quantity or property."""

    def __init__(self, network, view, func, args=None, kwargs=None):
        self.view = view
        self.net = network
        self.args = () if args is None else args
        self.kwargs = {} if args is None else kwargs
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.__class__(self.net, self.view, self.func, args=args, kwargs=kwargs)

    def __getitem__(self, idx):
        if idx not in self.view:
            raise IDNotFound(f'ID "{idx}" not in this view')
        return self.func(self.net, [idx], *self.args, **self.kwargs)[idx]

    def __repr__(self):
        cls = self.__class__.__name__
        fnc = self.func.__name__
        out = f"{cls}('{fnc}'"
        if self.args:
            out += f", args={self.args}"
        if self.kwargs:
            out += f", kwargs={self.kwargs}"
        out += ")"
        return out

    @property
    def name(self):
        """Name of this stat.

        The name of a stat is used to populate the keys of dictionaries in `MultiStat`
        objects, as well as the names of columns of pandas dataframes.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
        >>> da, d3 = H.nodes.degree, H.nodes.degree(order=3)
        >>> da.name, d3.name
        ('degree', 'degree(order=3)')
        >>> H.nodes.multi([da, d3]).asdict(transpose=True).keys()
        dict_keys(['degree', 'degree(order=3)'])
        >>> H.nodes.multi([da, d3]).aspandas().columns
        Index(['degree', 'degree(order=3)'], dtype='object')

        """
        name = f"{self.func.__name__}"
        if self.args or self.kwargs:
            args = [f"{s}" for s in self.args]
            kwargs = [f"{k}={v}" for k, v in self.kwargs.items()]
            name += "(" + ", ".join(args + kwargs) + ")"
        return name

    def __iter__(self):
        return iter(self._val.items())

    def items(self):
        return self._val.items()

    @property
    def _val(self):
        return self.func(self.net, self.view.ids, *self.args, **self.kwargs)

    def asdict(self):
        """Output the stat as a dict.

        Notes
        -----
        All stats are stored as dicts and therefore this method incurs in no overhead as
        type conversion is not necessary.

        """
        val = self._val
        return {n: val[n] for n in self.view}

    def aslist(self):
        """Output the stat as a list."""
        val = self._val
        return [val[n] for n in self.view]

    def asnumpy(self):
        """Output the stat as a numpy array."""
        return np.array(self.aslist())

    def aspandas(self):
        """Output the stat as a pandas series.

        Notes
        -----
        The `name` attribute of the returned series is set using the `name` property.

        """
        return pd.Series(self._val, name=self.name)

    def ashist(self, bins=10, bin_edges=False, density=False, log_binning=False):
        """Return the distribution of a numpy array.

        Parameters
        ----------
        vals : Numpy array
            The array of values
        bins : int, list, or Numpy array
            The number of bins or the bin edges.
        bin_edges : bool
            Whether to also output the min and max of each bin,
            by default, False.
        density : bool
            Whether to normalize the resulting distribution.
        log_binning : bool
            Whether to bin the values with log-sized bins.
            By default, False.


        Returns
        -------
        Pandas DataFrame
            A two-column table with "bin_center" and "value" columns,
            where "value" is a count or a probability. If `bin_edges`
            is True, outputs two additional columns, `bin_lo` and `bin_hi`,
            which outputs the left and right bin edges respectively.

        Notes
        -----
        Originally from https://github.com/jkbren/networks-and-dataviz

        """
        return hist(self.asnumpy(), bins, bin_edges, density, log_binning)

    def max(self):
        """The maximum value of this stat."""
        return self.asnumpy().max(axis=0)

    def min(self):
        """The minimum value of this stat."""
        return self.asnumpy().min(axis=0)

    def sum(self):
        """The sum of this stat."""
        return self.asnumpy().sum(axis=0)

    def mean(self):
        """The arithmetic mean of this stat."""
        return self.asnumpy().mean(axis=0)

    def median(self):
        """The median of this stat."""
        return np.median(self.asnumpy(), axis=0)

    def std(self):
        """The standard deviation of this stat."""
        return self.asnumpy().std(axis=0)

    def var(self):
        """The variance of this stat."""
        return self.asnumpy().var(axis=0)

    def moment(self, order=2, center=False):
        """The statistical moments of this stat.

        Parameters
        ----------
        order : int (default 2)
            The order of the moment.
        center : bool (default False)
            Whether to compute the centered (False) or uncentered/raw (True) moment.

        """
        arr = self.asnumpy()
        return spmoment(arr, moment=order) if center else np.mean(arr**order)


class NodeStat(IDStat):
    """An arbitrary node-quantity mapping.

    `NodeStat` objects represent a mapping that assigns a value to each node in a
    network.  For more details, see the `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """


class DiNodeStat(IDStat):
    """An arbitrary node-quantity mapping.

    `NodeStat` objects represent a mapping that assigns a value to each node in a
    network.  For more details, see the `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """


class EdgeStat(IDStat):
    """An arbitrary edge-quantity mapping.

    `EdgeStat` objects represent a mapping that assigns a value to each edge in a
    network.  For more details, see the `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """


class DiEdgeStat(IDStat):
    """An arbitrary edge-quantity mapping.

    `EdgeStat` objects represent a mapping that assigns a value to each edge in a
    network.  For more details, see the `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """


class MultiIDStat(IDStat):
    """Multiple mappings."""

    statsclass = None
    """IDStat subclass to use."""

    statsmodule = None
    """Module in which to search for mappings."""

    def __init__(self, network, view, stats):
        super().__init__(network, view, None)
        if isinstance(stats, self.statsclass):
            name = self.statsclass.__name__
            raise TypeError(f"must pass an iterable of {name}, not a single {name}")
        elif isinstance(stats, str):
            raise TypeError(
                f"must pass an iterable of {self.statsclass.__name__}, not str"
            )
        self.stats = [self._get_stat(f) for f in stats]

    def _get_stat(self, s):
        if isinstance(s, str):
            return self.statsclass(self.net, self.view, getattr(self.statsmodule, s))
        elif isinstance(s, self.statsclass):
            return s
        else:
            raise TypeError(f"{s.__name__} must be str or {self.statsclass.__name__}")

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            + "("
            + ", ".join(s.name for s in self.stats)
            + ")"
        )

    @property
    def name(self):
        return "[" + ", ".join(s.name for s in self.stats) + "]"

    @property
    def _val(self):
        result = {s.name: s.asdict() for s in self.stats}
        return {n: {s.name: result[s.name][n] for s in self.stats} for n in self.view}

    def asdict(self, inner=dict, transpose=False):
        """Output the stats as a dict of collections.

        Parameters
        ----------
        inner : dict (default) or list
            The type of the inner collections.  If dict (default), output a dict of
            dicts.  If list, output a dict of lists.
        transpose : bool (default False)
            By default, output a dict of dicts whose outer keys are the nodes and inner
            keys are the specified stats.  If True, the outer and inner keys are
            reversed.  Only used when `inner` is `dict`.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
        >>> m = H.nodes.multi(['degree', 'clustering_coefficient'])
        >>> m.asdict() # doctest: +NORMALIZE_WHITESPACE
        {1: {'degree': 1, 'clustering_coefficient': 1.0},
         2: {'degree': 2, 'clustering_coefficient': 0.6666666666666666},
         3: {'degree': 3, 'clustering_coefficient': 0.6666666666666666},
         4: {'degree': 2, 'clustering_coefficient': 1.0},
         5: {'degree': 2, 'clustering_coefficient': 1.0}}
        >>> m.asdict(transpose=True) # doctest: +NORMALIZE_WHITESPACE
        {'degree': {1: 1, 2: 2, 3: 3, 4: 2, 5: 2},
        'clustering_coefficient': {1: 1.0,
        2: 0.6666666666666666,
        3: 0.6666666666666666,
        4: 1.0,
        5: 1.0}}
        """
        val = self._val
        if inner is dict:
            if not transpose:
                return {n: val[n] for n in self.view}
            else:
                return {s.name: s.asdict() for s in self.stats}
        elif inner is list:
            return {n: list(val[n].values()) for n in self.view}
        else:
            raise ValueError

    def aslist(self, inner=list, transpose=False):
        """Output the stats as a list of collections.

        Parameters
        ----------
        inner : list (default) or dict
            The type of the inner collections.  If list (default), output a list of
            lists.  If dict, output a list of dicts.
        transpose : bool (default False)

            By default, output a list of lists where each inner list contains the stats
            of a single node.  If True, each inner list contains the values of a single
            stat of all nodes.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
        >>> m = H.nodes.multi(['degree', 'clustering_coefficient'])
        >>> m.aslist() # doctest:
        [[1, 1.0], [2, 0.6666666666666666], [3, 0.6666666666666666], [2, 1.0], [2, 1.0]]
        >>> m.aslist(transpose=True)
        [[1, 2, 3, 2, 2], [1.0, 0.6666666666666666, 0.6666666666666666, 1.0, 1.0]]
        """
        val = self._val
        if inner is list:
            if not transpose:
                return [list(val[n].values()) for n in self.view]
            else:
                return [s.aslist() for s in self.stats]
        elif inner is dict:
            return [val[n] for n in self.view]
        else:
            raise ValueError

    def asnumpy(self):
        """Output the stats as a numpy array.

        Notes
        -----
        Equivalent to `np.array(self.aslist(inner=list))`.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
        >>> H.nodes.multi(['degree', 'clustering_coefficient']).asnumpy()
        ... # doctest: +NORMALIZE_WHITESPACE
        array([[1.        , 1.        ],
               [2.        , 0.66666667],
               [3.        , 0.66666667],
               [2.        , 1.        ],
               [2.        , 1.        ]])

        """
        return np.array(self.aslist(inner=list))

    def aspandas(self):
        """Output the stats as a pandas dataframe.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph([[1, 2, 3], [2, 3, 4, 5], [3, 4, 5]])
        >>> H.nodes.multi(['degree', 'clustering_coefficient']).aspandas()
        ... # doctest: +NORMALIZE_WHITESPACE
           degree  clustering_coefficient
        1       1    1.000000
        2       2    0.666667
        3       3    0.666667
        4       2    1.000000
        5       2    1.000000

        """
        result = {s.name: s._val for s in self.stats}
        series = [pd.Series(v, name=k) for k, v in result.items()]
        return pd.concat(series, axis=1)

    def ashist(self, bins=10, bin_edges=False, density=False, log_binning=False):
        """Return the distributions of a numpy array.

        Parameters
        ----------
        vals : Numpy array
            The array of values
        bins : int, list, or Numpy array
            The number of bins or the bin edges.
        bin_edges : bool
            Whether to also output the min and max of each bin,
            by default, False.
        density : bool
            Whether to normalize the resulting distribution.
        log_binning : bool
            Whether to bin the values with log-sized bins.
            By default, False.


        Returns
        -------
        list of Pandas DataFrames
            Each entry of the list is a two-column table with "bin_center"
            and "value" columns, where "value" is a count or a probability.
            If `bin_edges` is True, outputs two additional columns,
            `bin_lo` and `bin_hi`, which outputs the left and right
            bin edges respectively.

        Notes
        -----
        Originally from https://github.com/jkbren/networks-and-dataviz

        """
        return [
            hist(data, bins, bin_edges, density, log_binning)
            for data in self.asnumpy().T
        ]


class MultiNodeStat(MultiIDStat):
    """Multiple node-quantity mappings.

    For more details, see the `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """

    statsclass = NodeStat
    statsmodule = nodestats


class MultiDiNodeStat(MultiIDStat):
    """Multiple node-quantity mappings.

    For more details, see the `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """

    statsclass = DiNodeStat
    statsmodule = dinodestats


class MultiEdgeStat(MultiIDStat):
    """Multiple edge-quantity mappings.

    For more details, see the `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """

    statsclass = EdgeStat
    statsmodule = edgestats


class MultiDiEdgeStat(MultiIDStat):
    """Multiple edge-quantity mappings.

    For more details, see the `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """

    statsclass = DiEdgeStat
    statsmodule = diedgestats


_dispatch_data = {
    "node": {
        "module": nodestats,
        "statclass": NodeStat,
        "multistatclass": MultiNodeStat,
    },
    "dinode": {
        "module": dinodestats,
        "statclass": DiNodeStat,
        "multistatclass": MultiDiNodeStat,
    },
    "edge": {
        "module": edgestats,
        "statclass": EdgeStat,
        "multistatclass": MultiEdgeStat,
    },
    "diedge": {
        "module": diedgestats,
        "statclass": DiEdgeStat,
        "multistatclass": MultiDiEdgeStat,
    },
}


def dispatch_stat(kind, net, view, name):
    try:
        func = getattr(_dispatch_data[kind]["module"], name)
    except AttributeError as e:
        raise AttributeError(f"Stat '{name}' not defined") from e
    return _dispatch_data[kind]["statclass"](net, view, func)


def dispatch_many_stats(kind, net, view, stats):
    return _dispatch_data[kind]["multistatclass"](net, view, stats)


def nodestat_func(func):
    """Decorate arbitrary functions to behave like :class:`NodeStat` objects.

    Parameters
    ----------
    func : callable
        Function or callable with signature `func(net, bunch)`, where `net` is the
        network and `bunch` is an iterable of nodes in `net`.  The call `func(net,
        bunch)` must return a dict with pairs of the form `(node: value)` where `node`
        is in `bunch` and `value` is the value of the statistic at `node`.

    Returns
    -------
    callable
        The decorated callable unmodified, after registering it in the `stats`
        framework.

    See Also
    --------
    :func:`edgestat_func`

    Notes
    -----
    The user must make sure that `func` is such that, if `res` is defined as `res =
    func(net, bunch)`, then `res` has keys in the same order as they are found in
    `bunch`.  Since python dicts preserve order, it is enough for `func` to create the
    returned dict by iterating over `bunch`.


    Examples
    --------
    >>> import xgi
    >>> H = xgi.Hypergraph([[1, 2], [3, 4], [4, 5, 6]])

    The following function defines a node-integer mapping.

    >>> def my_degree(net, bunch):
    ...     return {n: 10 * net.degree(n) for n in bunch}

    Node statistics can be called from the network or from the NodeView.

    >>> H.degree()
    {1: 1, 2: 1, 3: 1, 4: 2, 5: 1, 6: 1}
    >>> H.nodes.degree
    NodeStat('degree')

    However, `my_degree` is not recognized as a node statistic.

    >>> H.my_degree() # doctest: +ELLIPSIS
    Traceback (most recent call last):
    AttributeError:...

    >>> H.nodes.my_degree # doctest: +ELLIPSIS
    Traceback (most recent call last):
    AttributeError:...

    Use the `nodestat_func` decorator to turn `my_degree` into a valid stat.

    >>> original_my_degree = my_degree
    >>> my_degree = xgi.nodestat_func(my_degree)
    >>> H.my_degree()
    {1: 10, 2: 10, 3: 10, 4: 20, 5: 10, 6: 10}
    >>> H.nodes.my_degree
    NodeStat('my_degree')

    Now the entirety of the interface of stat objects is available.

    >>> H.nodes.filterby('my_degree', 20)
    NodeView((4,))
    >>> H.nodes.multi(['degree', 'my_degree']).aspandas()
       degree  my_degree
    1       1         10
    2       1         10
    3       1         10
    4       2         20
    5       1         10
    6       1         10

    Note the passed function is left unmodified.

    >>> my_degree is original_my_degree
    True

    The previous usage of `nodestat` is made for explanatory purposes.  A more typical
    use of `nodestat` is the following.

    >>> @xgi.nodestat_func
    ... def my_degree(net, bunch):
    ...     return {n: 10 * net.degree(n) for n in bunch}

    """
    setattr(nodestats, func.__name__, func)
    return func


def dinodestat_func(func):
    """Decorator that allows arbitrary functions to behave like :class:`DiNodeStat` objects.

    Works identically to :func:`nodestat`.  For extended documentation, see
    :func:`nodestat_func`.

    Parameters
    ----------
    func : callable
        Function or callable with signature `func(net, bunch)`, where `net` is the
        network and `bunch` is an iterable of edges in `net`.  The call `func(net,
        bunch)` must return a dict with pairs of the form `(edge: value)` where `edge`
        is in `bunch` and `value` is the value of the statistic at `edge`.

    Returns
    -------
    callable
        The decorated callable unmodified, after registering it in the `stats` framework.

    See Also
    --------
    :func:`nodestat_func`
    :func:`edgestat_func`
    :func:`diedgestat_func`

    """
    setattr(dinodestats, func.__name__, func)
    return func


def edgestat_func(func):
    """Decorator that allows arbitrary functions to behave like :class:`EdgeStat` objects.

    Works identically to :func:`nodestat`.  For extended documentation, see
    :func:`nodestat_func`.

    Parameters
    ----------
    func : callable
        Function or callable with signature `func(net, bunch)`, where `net` is the
        network and `bunch` is an iterable of edges in `net`.  The call `func(net,
        bunch)` must return a dict with pairs of the form `(edge: value)` where `edge`
        is in `bunch` and `value` is the value of the statistic at `edge`.

    Returns
    -------
    callable
        The decorated callable unmodified, after registering it in the `stats` framework.

    See Also
    --------
    :func:`nodestat_func`
    :func:`edgestat_func`
    :func:`diedgestat_func`

    """
    setattr(dinodestats, func.__name__, func)
    return func


def edgestat_func(func):
    """Decorate arbitrary functions to behave like :class:`EdgeStat` objects.

    Works identically to :func:`nodestat`.  For extended documentation, see
    :func:`nodestat_func`.

    Parameters
    ----------
    func : callable
        Function or callable with signature `func(net, bunch)`, where `net` is the
        network and `bunch` is an iterable of edges in `net`.  The call `func(net,
        bunch)` must return a dict with pairs of the form `(edge: value)` where `edge`
        is in `bunch` and `value` is the value of the statistic at `edge`.

    Returns
    -------
    callable
        The decorated callable unmodified, after registering it in the `stats`
        framework.

    See Also
    --------
    :func:`nodestat_func`

    """
    setattr(edgestats, func.__name__, func)
    return func


def diedgestat_func(func):
    """Decorator that allows arbitrary functions to behave like :class:`DiEdgeStat` objects.

    Works identically to :func:`nodestat`.  For extended documentation, see
    :func:`nodestat_func`.

    Parameters
    ----------
    func : callable
        Function or callable with signature `func(net, bunch)`, where `net` is the
        network and `bunch` is an iterable of edges in `net`.  The call `func(net,
        bunch)` must return a dict with pairs of the form `(edge: value)` where `edge`
        is in `bunch` and `value` is the value of the statistic at `edge`.

    Returns
    -------
    callable
        The decorated callable unmodified, after registering it in the `stats` framework.

    See Also
    --------
    :func:`nodestat_func`
    :func:`dinodestat_func`
    :func:`diedgestat_func`

    """
    setattr(diedgestats, func.__name__, func)
    return func



"""Generators for some classic hypergraphs.

All the functions in this module return a Hypergraph class (i.e. a simple, undirected
hypergraph).

"""

from itertools import chain, combinations

from ..utils import powerset

__all__ = [
    "empty_hypergraph",
    "empty_dihypergraph",
    "empty_simplicial_complex",
    "trivial_hypergraph",
    "complete_hypergraph",
    "complement",
]


def _empty_network(create_using, default):
    """Return an empty network.

    See Also
    --------
    empty_hypergraph
    empty_simplicial_complex

    """
    if create_using is None:
        H = default()
    elif hasattr(create_using, "_node") or hasattr(create_using, "_node_in"):
        # create_using is a Hypergraph object
        create_using.clear()
        H = create_using
    else:
        # try create_using as constructor
        H = create_using()
    return H


def empty_hypergraph(create_using=None, default=None):
    """Returns the empty hypergraph with zero nodes and edges.

    Parameters
    ----------
    create_using : Hypergraph Instance, Constructor or None
        If None, use the `default` constructor.
        If a constructor, call it to create an empty hypergraph.
    default : Hypergraph constructor (default None)
        The constructor to use if create_using is None.
        If None, then xgi.Hypergraph is used.

    Returns
    -------
    Hypergraph object
        An empty hypergraph

    See also
    --------
    empty_simplicial_complex
    trivial_hypergraph

    Examples
    --------
    >>> import xgi
    >>> H = xgi.empty_hypergraph()
    >>> H.num_nodes, H.num_edges
    (0, 0)

    """
    # this import needs to happen when the function runs, not when the module is first
    # imported, to avoid circular imports
    import xgi

    if default is None:
        default = xgi.Hypergraph
    return _empty_network(create_using, default)


def empty_dihypergraph(create_using=None, default=None):
    """Returns the empty dihypergraph with zero nodes and edges.

    Parameters
    ----------
    create_using : DiHypergraph Instance, Constructor or None
        If None, use the `default` constructor.
        If a constructor, call it to create an empty dihypergraph.
    default : DiHypergraph constructor (default None)
        The constructor to use if create_using is None.
        If None, then xgi.Hypergraph is used.

    Returns
    -------
    Hypergraph object
        An empty hypergraph

    See also
    --------
    empty_simplicial_complex
    trivial_hypergraph

    Examples
    --------
    >>> import xgi
    >>> H = xgi.empty_dihypergraph()
    >>> H.num_nodes, H.num_edges
    (0, 0)

    """
    # this import needs to happen when the function runs, not when the module is first
    # imported, to avoid circular imports
    import xgi

    if default is None:
        default = xgi.DiHypergraph
    return _empty_network(create_using, default)


def empty_simplicial_complex(create_using=None, default=None):
    """Returns the empty simplicial complex with zero nodes and simplices.

    Parameters
    ----------
    create_using : SimplicialComplex Instance, Constructor or None
        If None, use the `default` constructor.
        If a constructor, call it to create an empty simplicial complex.
    default : SimplicialComplex constructor (default None)
        The constructor to use if create_using is None.
        If None, then xgi.SimplicialComplex is used.

    Returns
    -------
    SimplicialComplex
        An empty simplicial complex.

    See also
    --------
    empty_hypergraph
    trivial_hypergraph

    Examples
    --------
    >>> import xgi
    >>> H = xgi.empty_simplicial_complex()
    >>> H.num_nodes, H.num_edges
    (0, 0)

    """
    # this import needs to happen when the function runs, not when the module is first
    # imported, to avoid circular imports
    import xgi

    if default is None:
        default = xgi.SimplicialComplex
    return _empty_network(create_using, default)


def trivial_hypergraph(n=1, create_using=None, default=None):
    """Returns a hypergraph with `n` nodes and zero edges.

    Parameters
    ----------
    n : int, optional
        Number of nodes (default is 1)
    create_using : Hypergraph Instance, Constructor or None
        If None, use the `default` constructor.
        If a constructor, call it to create an empty hypergraph.
    default : Hypergraph constructor (default None)
        The constructor to use if create_using is None.
        If None, then xgi.Hypergraph is used.

    Returns
    -------
    Hypergraph object
        A trivial hypergraph with `n` nodes

    See also
    --------
    empty_hypergraph
    empty_simplicial_complex

    Examples
    --------
    >>> import xgi
    >>> H = xgi.trivial_hypergraph()
    >>> H.num_nodes, H.num_edges
    (1, 0)

    """
    # this import needs to happen when the function runs, not when the module is first
    # imported, to avoid circular imports
    import xgi

    if default is None:
        default = xgi.Hypergraph
    H = _empty_network(create_using, default)
    nodes = range(n)
    H.add_nodes_from(nodes)

    return H


def complete_hypergraph(N, order=None, max_order=None, include_singletons=False):
    """Generate a complete hypergraph, i.e. one that contains all possible hyperdges
    at a given `order` or up to a `max_order`.

    Parameters
    ----------

    N : int
        Number of nodes
    order : int or None
        If not None (default), specifies the single order for which to generate
        hyperedges
    max_order : int or None
        If not None (default), specifies the maximum order for which to generate
        hyperedges
    include_singletons : bool
        Whether to include singleton edges (default: False). This argument is discarded
        if max_order is None.

    Return
    ------
    Hypergraph object
        A complete hypergraph with `N` nodes

    Notes
    ----
    Only one of `order` and `max_order` can be specified by and int (not None).
    Additionally, at least one of either must be specified.

    The number of possible edges grows exponentially as :math:`2^N` for large `N` and
    quickly becomes impractically long to compute, especially when using
    `max_order`. For example, `N=100` and `max_order=5` already yields :math:`10^8`
    edges. Increasing `N=1000` makes it :math:`10^{13}`. `N=100` and with a larger
    `max_order=6` yields :math:`10^9` edges.

    """
    # this import needs to happen when the function runs, not when the module is first
    # imported, to avoid circular imports
    import xgi

    if (order is None) and (max_order is None):
        raise ValueError(
            "At least one among order and max_order must be specified (not None)"
        )
    if (order is not None) and (max_order is not None):
        raise ValueError(
            "Both order and max_order cannot be specified (not None) at the same time."
        )

    H = xgi.Hypergraph()

    nodes = range(N)
    H.add_nodes_from(nodes)

    if order is not None:
        edges = combinations(nodes, order + 1)
    elif max_order is not None:
        start = 1 if include_singletons else 2
        end = max_order + 1
        assert end >= start  # can be equal because adding +1 to end below

        s = list(nodes)
        edges = chain.from_iterable(combinations(s, r) for r in range(start, end + 1))

    H.add_edges_from(edges)

    return H


def complement(H):
    """Returns the complement of hypergraph H.

    The complement Hc of a hypergraph H has the same nodes (same indexes) as H and
    contains every possible hyperedge linking these nodes, except those present in H.
    Also, the maximum hyperedge size in Hc is the same as in H.
    Hyperedges of size one are taken into account.

    Parameters
    ----------
    H : xgi.Hypergraph
        Hypergraph to complement.

    Returns
    -------
    Hc : xgi.Hypergraph
        Complement of H.
    """

    from ..algorithms import max_edge_order

    N = len(H.nodes)
    M = len(H.edges)

    if N == 0:
        return empty_hypergraph()
    else:

        node_dict = dict(zip(H.nodes, range(N)))
        num_dict = {idx: n for n, idx in node_dict.items()}

        # parsing list of edges into set of strings
        edges = set()
        for st in H.edges.members():
            lst = [node_dict[n] for n in st]  # we let go node IDs
            lst.sort()
            string = ",".join(str(idx) for idx in lst)
            edges.add(string)

        # parsing list of possible edges into set of strings
        max_edge_size = max_edge_order(H) + 1
        possible_edges = set()
        iterator = powerset(
            range(N),
            include_empty=False,
            include_singletons=True,
            max_size=max_edge_size,
        )
        for subset in iterator:
            lst = list(subset)
            lst.sort()
            string = ",".join(str(idx) for idx in lst)
            possible_edges.add(string)

        # performing set difference
        to_add = possible_edges.difference(edges)

        Hc = empty_hypergraph()
        Hc.add_nodes_from(H.nodes)

        for string in to_add:
            lst = string.split(",")
            nodes = [num_dict[int(idx)] for idx in lst]
            Hc.add_edge(nodes)

        return Hc



"""Algorithms for computing shortest paths in a hypergraph."""

import numpy as np

from ..utils import utilities

__all__ = ["single_source_shortest_path_length", "shortest_path_length"]


def single_source_shortest_path_length(H, source):
    """
    Returns the distances from source to every other node in hypergraph H.

    Parameters
    ----------
    H : xgi.Hypergraph
        Hypergraph on which to compute the distances. Node indexes must be integers.
    source : int
        Index of the node from which to compute the distance to every other node.

    Returns
    -------
    dists : dict
        Dictionary where keys are node indexes and values are the distances from source.
    """

    # 1. Mark all nodes unvisited.
    is_unseen = dict()
    for node in H.nodes:
        is_unseen[node] = True
    n_unseen = len(H.nodes)

    # 2. Assign to every node a tentative distance value.
    dists = dict()
    for node in H.nodes:
        dists[node] = np.infty
    dists[source] = 0
    is_unseen[source] = False
    n_unseen -= 1
    current = source

    # 3. Consider all of unvisited neighbors of current node and calculate their tentative distances to the current node.
    stop_condition = False
    while not stop_condition:
        for ngb in H.nodes.neighbors(current):
            if is_unseen[ngb]:
                increment = 1  # increment = weight(ngb, current) if weighted
                new_dist = dists[current] + increment
                if new_dist < dists[ngb]:
                    dists[ngb] = new_dist
                else:
                    pass
            else:
                pass

        # 4. Mark the current node as visited and remove it from the unvisited set.
        is_unseen[current] = False
        n_unseen -= 1

        # 5. Check for stop condition.
        stop_condition = (n_unseen == 0) or (
            utilities.min_where(dists, is_unseen) == np.infty
        )

        # 6. Otherwise, select the unvisited node that is marked with the smallest tentative distance, set it as the new current node, and go back to step 3.
        min_val = np.infty
        argmin = current
        for node in dists.keys():
            if is_unseen[node]:
                if dists[node] < min_val:
                    min_val = dists[node]
                    argmin = node
                else:
                    pass
            else:
                pass
        current = argmin

    return dists


def shortest_path_length(H):
    """
    Returns a generator of tuples (source, dists) where dists is a dictonary
    containing the distances from source to every other node in hypergraph H,
    for all possible source in H.

    Parameters
    ----------
    H : xgi.Hypergraph
        Hypergraph on which to compute the distances. Node indexes must be integers.

    Returns
    -------
    paths : generator of tuples
        Every tuple is of the form (source, dict_of_lengths), for every possible source.
    """

    for source in H.nodes:
        dists = single_source_shortest_path_length(H, source)
        yield (source, dists)



"""Laplacian matrices associated to hypergraphs.

Note that the order of the rows and columns of the
matrices in this module correspond to the order in
which nodes/edges are added to the hypergraph or
simplicial complex. If the node and edge IDs are
able to be sorted, the following is an example to sort
by the node and edge IDs.

>>> import xgi
>>> import pandas as pd
>>> H = xgi.Hypergraph([[1, 2, 3, 7], [4], [5, 6, 7]])
>>> I, nodedict, edgedict = xgi.incidence_matrix(H, sparse=False, index=True)
>>> # Sorting the resulting numpy array:
>>> sortedI = I.copy()
>>> sortedI = sortedI[sorted(nodedict, key=nodedict.get), :]
>>> sortedI = sortedI[:, sorted(edgedict, key=edgedict.get)]
>>> sortedI
array([[1, 0, 0],
       [1, 0, 0],
       [1, 0, 0],
       [0, 1, 0],
       [0, 0, 1],
       [0, 0, 1],
       [1, 0, 1]])
>>> # Indexing a Pandas dataframe by the node/edge IDs
>>> df = pd.DataFrame(I, index=nodedict.values(), columns=edgedict.values())

If the nodes are already sorted, this order can be preserved by adding the nodes
to the hypergraph prior to adding edges. For example,

>>> import xgi
>>> H = xgi.Hypergraph()
>>> H.add_nodes_from(range(1, 8))
>>> H.add_edges_from([[1, 2, 3, 7], [4], [5, 6, 7]])
>>> xgi.incidence_matrix(H, sparse=False)
array([[1, 0, 0],
       [1, 0, 0],
       [1, 0, 0],
       [0, 1, 0],
       [0, 0, 1],
       [0, 0, 1],
       [1, 0, 1]])

"""
from warnings import warn

import numpy as np
from scipy.sparse import csr_array, diags

from ..exception import XGIError
from .hypergraph_matrix import adjacency_matrix, clique_motif_matrix, degree_matrix

__all__ = [
    "laplacian",
    "multiorder_laplacian",
    "normalized_hypergraph_laplacian",
]


def laplacian(H, order=1, sparse=False, rescale_per_node=False, index=False):
    """Laplacian matrix of order d, see [1].

    Parameters
    ----------
    HG : Hypergraph
        Hypergraph
    order : int
        Order of interactions to consider. If order=1 (default),
        returns the usual graph Laplacian.
    sparse: bool, default: False
        Specifies whether the output matrix is a scipy sparse matrix or a numpy matrix.
    index: bool, default: False
        Specifies whether to output disctionaries mapping the node and edge IDs to
        indices.

    Returns
    -------
    L_d : numpy array
        Array of dim (N, N)
    if index is True:
        return rowdict

    See also
    --------
    multiorder_laplacian

    References
    ----------
    .. [1] Lucas, M., Cencetti, G., & Battiston, F. (2020).
        Multiorder Laplacian for synchronization in higher-order networks.
        Physical Review Research, 2(3), 033410.

    """
    A, row_dict = adjacency_matrix(
        H, order=order, sparse=sparse, weighted=True, index=True
    )

    if A.shape == (0, 0):
        L = csr_array((0, 0)) if sparse else np.empty((0, 0))
        return (L, {}) if index else L

    if sparse:
        K = csr_array(diags(degree_matrix(H, order=order)))
    else:
        K = np.diag(degree_matrix(H, order=order))

    L = order * K - A  # ravel needed to convert sparse matrix

    if rescale_per_node:
        L = L / order

    return (L, row_dict) if index else L


def multiorder_laplacian(
    H, orders, weights, sparse=False, rescale_per_node=False, index=False
):
    """Multiorder Laplacian matrix, see [1].

    Parameters
    ----------
    HG : Hypergraph
        Hypergraph
    orders : list of int
        Orders of interactions to consider.
    weights: list of float
        Weights associated to each order, i.e coupling strengths gamma_i in [1].
    sparse: bool, default: False
        Specifies whether the output matrix is a scipy sparse matrix or a numpy matrix
    rescale_per_node: bool, (default=False)
        Whether to rescale each Laplacian of order d by d (per node).
    index: bool, default: False
        Specifies whether to output dictionaries mapping the node and edge IDs to
        indices.

    Returns
    -------
    L_multi : numpy array
        Array of dim (N, N)
    if index is True:
        return rowdict

    See also
    --------
    laplacian

    References
    ----------
    .. [1] Lucas, M., Cencetti, G., & Battiston, F. (2020).
        Multiorder Laplacian for synchronization in higher-order networks.
        Physical Review Research, 2(3), 033410.

    """
    if len(orders) != len(weights):
        raise ValueError("orders and weights must have the same length.")

    Ls = [
        laplacian(H, order=d, sparse=sparse, rescale_per_node=rescale_per_node)
        for d in orders
    ]
    Ks = [degree_matrix(H, order=d) for d in orders]

    if sparse:
        L_multi = csr_array((H.num_nodes, H.num_nodes))
    else:
        L_multi = np.zeros((H.num_nodes, H.num_nodes))

    for L, K, w, d in zip(Ls, Ks, weights, orders):
        if np.all(K == 0):
            # avoid getting nans from dividing by 0
            # manually setting contribution to 0 as it should be
            warn(
                f"No edges of order {d}. Contribution of "
                "that order is zero. Its weight is effectively zero."
            )
        else:
            L_multi += L * w / np.mean(K)

    rowdict = {i: v for i, v in enumerate(H.nodes)}

    return (L_multi, rowdict) if index else L_multi


def normalized_hypergraph_laplacian(H, sparse=True, index=False):
    """Compute the normalized Laplacian.

    Parameters
    ----------
    H : Hypergraph
        Hypergraph
    sparse : bool, optional
        whether or not the laplacian is sparse, by default True
    index : bool, optional
        whether to return a dictionary mapping IDs to rows, by default False

    Returns
    -------
    array
        csr_array if sparse and if not, a numpy ndarray
    dict
        a dictionary mapping node IDs to rows and columns
        if index is True.


    Raises
    ------
    XGIError
        If there are isolated nodes.

    References
    ----------
    "Learning with Hypergraphs: Clustering, Classification, and Embedding"
    by Dengyong Zhou, Jiayuan Huang, Bernhard Schölkopf
    Advances in Neural Information Processing Systems (2006)

    """
    if H.nodes.isolates():
        raise XGIError(
            "Every node must be a member of an edge to avoid divide by zero error!"
        )

    D = degree_matrix(H)
    A, rowdict = clique_motif_matrix(H, sparse=sparse, index=True)

    if sparse:
        Dinvsqrt = csr_array(diags(np.power(D, -0.5)))
        eye = csr_array((H.num_nodes, H.num_nodes))
        eye.setdiag(1)
    else:
        Dinvsqrt = np.diag(np.power(D, -0.5))
        eye = np.eye(H.num_nodes)

    L = 0.5 * (eye - Dinvsqrt @ A @ Dinvsqrt)
    return (L, rowdict) if index else L



import networkx as nx

from ..exception import XGIError

__all__ = ["to_encapsulation_dag", "empirical_subsets_filter"]


def to_encapsulation_dag(H, subset_types="all"):
    """The encapsulation DAG (Directed Acyclic Graph) of
    the hypergraph H.

    An encapsulation DAG is a directed line graph
    where the nodes are hyperedges in H and a directed edge
    exists from a larger hyperedge to a smaller hyperedge if
    the smaller is a subset of the larger.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest
    subset_types : str, optional
        Type of relations to include. Options are:
            "all" : all subset relationships
            "immediate" : only subset relationships between hyperedges of
                adjacent sizes (i.e., edges from k to k-1)
            "empirical" : A relaxation of the "immediate" option where only
                subset relationships between hyperedges of size k and subsets
                of maximum size k'<k existing in the hypergraph are included.
                For example, a hyperedge of size 5 may have no immediate
                encapsulation relationships with hyperedges of size 4, but may
                encapsulate hyperedegs of size 3, which will be included if
                using this setting (whereas relationships with subsets of size 2
                would not be included).
    Returns
    -------
    LG : networkx.DiGraph
         The line graph associated to the Hypergraph

    Examples
    --------
    >>> import xgi
    >>> from xgi.convert import to_encapsulation_dag, empirical_subsets_filter
    >>> H = xgi.Hypergraph([["a","b","c"], ["b","c","f"], ["a","b"], ["c", "e"], ["a"], ["f"]])
    >>> dag = to_encapsulation_dag(H)
    >>> dag.edges()
    OutEdgeView([(0, 2), (0, 4), (2, 4), (1, 5)])
    >>> dag = to_encapsulation_dag(H, subset_types="immediate")
    >>> dag.edges()
    OutEdgeView([(0, 2), (2, 4)])
    >>> dag = to_encapsulation_dag(H, subset_types="empirical")
    >>> dag.edges()
    OutEdgeView([(0, 2), (2, 4), (1, 5)])

    References
    ----------
    "Encapsulation Structure and Dynamics in Hypergraphs", by Timothy LaRock
    & Renaud Lambiotte. https://arxiv.org/abs/2307.04613

    """
    if not (subset_types in ["all", "immediate", "empirical"]):
        raise XGIError(
            f"{subset_types} not a valid subset_types option. Choices are "
            "'all', 'immediate', and 'empirical'."
        )

    # Construct the dag
    dag = nx.DiGraph()
    # Loop over hyperedges
    for he_idx in H.edges:
        # Add the hyperedge as a node
        dag.add_node(he_idx)
        # Get the hyperedge as a set
        he = H.edges.members(he_idx)
        # Get candidate encapsulation hyperedges
        candidates = _get_candidates(subset_types, H, he)
        # for each candidate
        candidates_checked = set()
        for cand_idx in candidates:
            if cand_idx in candidates_checked:
                continue
            cand = H.edges.members(cand_idx)
            if len(he) > len(cand):
                if _encapsulated(he, cand):
                    dag.add_edge(he_idx, cand_idx)
            elif len(cand) > len(he):
                if _encapsulated(cand, he):
                    dag.add_edge(cand_idx, he_idx)

    # If empirically closest subsets, filter out all edges except those
    # between k and maximum existing k'<k
    if subset_types == "empirical":
        empirical_subsets_filter(H, dag)

    return dag


def _encapsulated(larger, smaller):
    """Test if a larger hyperedge encapsulates a smaller hyperedge.

    Parameters
    ----------
    larger : Set
        larger hyperedge
    smaller : Set
        smaller hyperedge
    Returns
    -------
    _ : Bool
        True if the size of the intersection between larger and smaller is
            the same as the size of smaller.

    """
    return len(set(larger).intersection(set(smaller))) == len(smaller)


def _get_candidates(subset_types, H, he):
    """Get the candidate hyperedges for encapsulation by he based on the subset type.

    Parameters
    ----------
    subset_types : str
        Type of subset relationships
    H : Hypergraph
        The hypergraph
    he : Set
        The hyperedge
    Returns
    -------
    candidates : Set
        A set of hyperedge IDs to check for encapsulation by hyperedge he

    """
    candidates = set()
    for node in he:
        for cand_idx in H.nodes.memberships(node):
            if cand_idx not in candidates:
                cand = H.edges.members(cand_idx)
                if _check_candidate(subset_types, he, cand):
                    candidates.add(cand_idx)

    return candidates


def _check_candidate(subset_types, he, cand):
    """Check whether a hyperedge cand is a candidate to be encapsulated by
    hyperedge he based on subset_types.

    Parameters
    ----------
    subset_types : str
        Type of subset relationships
    he : Set
        The hyperedge
    cand : Set
        The candidate
    Returns
    -------
    is_candidate : bool
        True if cand is a valid candidate to be encapsulated by he, False
        otherwise.
    """
    is_candidate = False
    if subset_types in ["all", "empirical"]:
        if len(he) != len(cand):
            is_candidate = True
    elif subset_types == "immediate":
        if len(he) == len(cand) - 1 or len(he) - 1 == len(cand):
            is_candidate = True
    return is_candidate


def empirical_subsets_filter(H, dag):
    """
    Filters encapsulation DAG of H in place to only include edges between hyperedges
    of size k and the maximum existing k'.

    Parameters
    ----------
    H : Hypergraph
        The hypergraph of interest
    dag : nx.DiGraph
        The encapsulation dag of H constructed with to_encapsulation_dag(H, subset_types="all")

    Returns
    -------
    dag : networkx.DiGraph
         The filtered line graph (also modified in place)

    References
    ----------
    "Encapsulation Structure and Dynamics in Hypergraphs", by Timothy LaRock
    & Renaud Lambiotte. https://arxiv.org/abs/2307.04613

    """
    # Loop over all edges
    for edge_idx in dag:
        preds = list(dag.predecessors(edge_idx))
        if len(preds) > 0:
            # Get the minimum superface size
            min_sup_size = min([len(H.edges.members(cand_idx)) for cand_idx in preds])
            # Keep only the superfaces with that size
            to_remove = []
            for cand_idx in preds:
                if len(H.edges.members(cand_idx)) != min_sup_size:
                    dag.remove_edge(cand_idx, edge_idx)

        # Repeat for subsets
        outs = list(dag.successors(edge_idx))
        if len(outs) > 0:
            max_sub_size = max([len(H.edges.members(sub_idx)) for sub_idx in outs])
            for cand_idx in outs:
                if len(H.edges.members(cand_idx)) != max_sub_size:
                    dag.remove_edge(edge_idx, cand_idx)
    return dag



"""Methods for converting to and from bipartite edgelists."""

from ..generators import empty_hypergraph

__all__ = ["from_bipartite_edgelist", "to_bipartite_edgelist"]


def from_bipartite_edgelist(edges, create_using=None):
    """Generate a hypergraph from a list of lists.

    Parameters
    ----------
    e : tuple, list, or array of tuples, lists, or arrays, each of size 2
        A bipartite edgelist
    create_using : Hypergraph constructor, optional
        The hypergraph to add the edges to, by default None

    Returns
    -------
    Hypergraph object
        The constructed hypergraph object

    See Also
    --------
    to_hyperedge_list
    """
    H = empty_hypergraph(create_using)
    for n, e in edges:
        H.add_node_to_edge(e, n)
    return H


def to_bipartite_edgelist(H):
    """Generate a hyperedge list from a hypergraph.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest

    Returns
    -------
    list of sets
        The hyperedge list

    See Also
    --------
    from_hyperedge_list
    """
    return [(n, id) for id, e in H.edges.members(dtype=dict).items() for n in e]



"""Methods for simplicial complexes."""

from ..core import Hypergraph, SimplicialComplex
from ..exception import XGIError
from ..generators import empty_simplicial_complex

__all__ = ["from_simplex_dict", "from_max_simplices"]


def from_simplex_dict(d, create_using=None):
    """Creates a Simplicial Complex from a dictionary of simplices,
    if the subfaces of existing simplices are not given in the dict
    then the function adds them with integer IDs.


    Parameters
    ----------
    d : dict
        A dictionary where the keys are simplex IDs and the values
        are containers of nodes specifying the simplices.
    create_using : SimplicialComplex constructor, optional
        The simplicial complex object to add the data to, by default None

    Returns
    -------
    SimplicialComplex object
        The constructed simplicial complex object

    """
    SC = empty_simplicial_complex(create_using)
    SC.add_simplices_from((members, uid) for uid, members in d.items())
    return SC


def from_max_simplices(SC):
    """Returns a hypergraph constructed from the
    maximal simplices of the provided simplicial complex.

    Parameters
    ----------
    SC : SimplicialComplex

    Returns
    -------
    Hypergraph

    """
    if type(SC) != SimplicialComplex:
        raise XGIError("The input must be a SimplicialComplex")

    max_simplices = SC.edges.maximal()
    H = Hypergraph()
    H.add_nodes_from(SC.nodes)  # to keep node order and isolated nodes
    H.add_edges_from([list(SC.edges.members(e)) for e in max_simplices])
    return H



"""Algorithms for computing nodal clustering coefficients."""

import numpy as np

from ..exception import XGIError
from ..linalg import adjacency_matrix

__all__ = [
    "clustering_coefficient",
    "local_clustering_coefficient",
    "two_node_clustering_coefficient",
]


def clustering_coefficient(H):
    r"""Return the clustering coefficients for
    each node in a Hypergraph.

    This clustering coefficient is defined as the
    clustering coefficient of the unweighted pairwise
    projection of the hypergraph, i.e.,
    :math:`c = A^3_{i,i}/\binom{k}{2},`
    where :math:`A` is the adjacency matrix of the network
    and :math:`k` is the pairwise degree of :math:`i`.

    Parameters
    ----------
    H : Hypergraph
        Hypergraph

    Returns
    -------
    dict
        nodes are keys, clustering coefficients are values.

    Notes
    -----
    The clustering coefficient is undefined when the number of
    neighbors is 0 or 1, but we set the clustering coefficient
    to 0 in these cases. For more discussion, see
    https://arxiv.org/abs/0802.2512

    See Also
    --------
    local_clustering_coefficient
    two_node_clustering_coefficient

    References
    ----------
    "Clustering Coefficients in Protein Interaction Hypernetworks"
    by Suzanne Gallagher and Debra Goldberg.
    DOI: 10.1145/2506583.2506635

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(3, [1, 1])
    >>> cc = xgi.clustering_coefficient(H)
    >>> cc
    {0: 1.0, 1: 1.0, 2: 1.0}
    """
    adj, index = adjacency_matrix(H, index=True)
    ndict = {n: i for i, n in index.items()}

    k = adj.sum(axis=1)
    denom = k * (k - 1) / 2
    mat = adj.dot(adj).dot(adj)

    with np.errstate(divide="ignore", invalid="ignore"):
        result = np.nan_to_num(0.5 * mat.diagonal() / denom)

    return {n: result[ndict[n]] if n in ndict else 0 for n in H.nodes}


def local_clustering_coefficient(H):
    """Compute the local clustering coefficient.

    This clustering coefficient is based on the
    overlap of the edges connected to a given node,
    normalized by the size of the node's neighborhood.

    Parameters
    ----------
    H : Hypergraph
        Hypergraph

    Returns
    -------
    dict
        keys are node IDs and values are the
        clustering coefficients.

    Notes
    -----
    The clustering coefficient is undefined when the number of
    neighbors is 0 or 1, but we set the clustering coefficient
    to 0 in these cases. For more discussion, see
    https://arxiv.org/abs/0802.2512

    See Also
    --------
    clustering_coefficient
    two_node_clustering_coefficient

    References
    ----------
    "Properties of metabolic graphs: biological organization or representation
    artifacts?"  by Wanding Zhou and Luay Nakhleh.
    https://doi.org/10.1186/1471-2105-12-132

    "Hypergraphs for predicting essential genes using multiprotein complex data"
    by Florian Klimm, Charlotte M. Deane, and Gesine Reinert.
    https://doi.org/10.1093/comnet/cnaa028

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(3, [1, 1])
    >>> cc = xgi.local_clustering_coefficient(H)
    >>> cc
    {0: 1.0, 1: 1.0, 2: 1.0}

    """
    result = {}

    memberships = H.nodes.memberships()
    members = H.edges.members()

    for n in H.nodes:
        ev = list(memberships[n])
        dv = len(ev)
        if dv <= 1:
            result[n] = 0
        else:
            total_eo = 0
            # go over all pairs of edges pairwise
            for e1 in range(dv):
                edge1 = members[e1]
                for e2 in range(e1):
                    edge2 = members[e2]
                    # set differences for the hyperedges
                    D1 = set(edge1) - set(edge2)
                    D2 = set(edge2) - set(edge1)
                    # if edges are the same by definition the extra overlap is zero
                    if len(D1.union(D2)) == 0:
                        eo = 0
                    else:
                        # otherwise we have to look at their neighbours
                        # the neighbours of D1 and D2, respectively.
                        neighD1 = {i for d in D1 for i in H.nodes.neighbors(d)}
                        neighD2 = {i for d in D2 for i in H.nodes.neighbors(d)}
                        # compute extra overlap [len() is used for cardinality of edges]
                        eo = (
                            len(neighD1.intersection(D2))
                            + len(neighD2.intersection(D1))
                        ) / len(
                            D1.union(D2)
                        )  # add it up
                    # add it up
                    total_eo = total_eo + eo

            # include normalisation by degree k*(k-1)/2
            result[n] = 2 * total_eo / (dv * (dv - 1))
    return result


def two_node_clustering_coefficient(H, kind="union"):
    """Return the clustering coefficients for
    each node in a Hypergraph.

    This definition averages over all of the
    two-node clustering coefficients involving the node.

    Parameters
    ----------
    H : Hypergraph
        Hypergraph
    kind : string, optional
        The type of two node clustering coefficient. Options
        are "union", "max", and "min". By default, "union".

    Returns
    -------
    dict
        nodes are keys, clustering coefficients are values.

    Notes
    -----
    The clustering coefficient is undefined when the number of
    neighbors is 0 or 1, but we set the clustering coefficient
    to 0 in these cases. For more discussion, see
    https://arxiv.org/abs/0802.2512

    See Also
    --------
    clustering_coefficient
    local_clustering_coefficient

    References
    ----------
    "Clustering Coefficients in Protein Interaction Hypernetworks"
    by Suzanne Gallagher and Debra Goldberg.
    DOI: 10.1145/2506583.2506635

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(3, [1, 1])
    >>> cc = xgi.two_node_clustering_coefficient(H, kind="union")
    >>> cc
    {0: 0.5, 1: 0.5, 2: 0.5}
    """
    result = {}
    memberships = H.nodes.memberships()
    for n in H.nodes:
        neighbors = H.nodes.neighbors(n)
        result[n] = 0.0
        for v in neighbors:
            result[n] += _uv_cc(n, v, memberships, kind=kind) / len(neighbors)
    return result


def _uv_cc(u, v, memberships, kind="union"):
    """Helper function to compute the two-node
    clustering coefficient.

    Parameters
    ----------
    u : hashable
        First node
    v : hashable
        Second node
    memberships : dict
        node IDs are keys, edge IDs to which they belong
        are values.
    kind : str, optional
        Type of clustering coefficient to compute, by default "union".
        Options:

        - "union"
        - "max"
        - "min"

    Returns
    -------
    float
        The clustering coefficient

    Raises
    ------
    XGIError
        If an invalid clustering coefficient kind
        is specified.

    References
    ----------
    "Clustering Coefficients in Protein Interaction Hypernetworks"
    by Suzanne Gallagher and Debra Goldberg.
    DOI: 10.1145/2506583.2506635
    """
    m_u = memberships[u]
    m_v = memberships[v]

    num = len(m_u.intersection(m_v))

    if kind == "union":
        denom = len(m_u.union(m_v))
    elif kind == "min":
        denom = min(len(m_u), len(m_v))
    elif kind == "max":
        denom = max(len(m_u), len(m_v))
    else:
        raise XGIError("Invalid kind of clustering.")

    if denom == 0:
        return np.nan

    return num / denom



"""View classes for dihypergraphs.

A View class allows for inspection and querying of an underlying object but does not
allow modification.  This module provides View classes for nodes and edges of a dihypergraph.
Views are automatically updaed when the dihypergraph changes.

"""

from collections.abc import Mapping, Set

from ..exception import IDNotFound, XGIError
from ..stats import IDStat, dispatch_many_stats, dispatch_stat

__all__ = [
    "DiNodeView",
    "DiEdgeView",
]


class DiIDView(Mapping, Set):
    """Base View class for accessing the ids (nodes or edges) of a DiHypergraph.

    Can optionally keep track of a subset of ids.  By default all node ids or all edge
    ids are kept track of.

    Parameters
    ----------
    id_dict : dict
        The original dict this is a view of.
    id_attrs : dict
        The original attribute dict this is a view of.
    ids : iterable
        A subset of the keys in id_dict to keep track of.

    Raises
    ------
    XGIError
        If ids is not a subset of the keys of id_dict.

    """

    _id_kind = None

    __slots__ = (
        "_net",
        "_ids",
    )

    def __getstate__(self):
        """Function that allows pickling.

        Returns
        -------
        dict
            The keys access the IDs and their attributes respectively
            and the values are dictionarys from the Hypergraph class.

        """
        return {
            "_net": self._net,
            "_ids": self._ids,
        }

    def __setstate__(self, state):
        """Function that allows unpickling.

        Parameters
        ----------
        dict
            The keys access the IDs and their attributes respectively
            and the values are dictionarys from the Hypergraph class.

        """
        self._net = state["_net"]
        self._id_kind = state["_id_kind"]

    def __init__(self, network, ids=None):
        self._net = network

        if self._id_kind == "dinode":
            self._in_id_dict = None if self._net is None else network._node_in
            self._out_id_dict = None if self._net is None else network._node_out
            self._id_attr = None if self._net is None else network._node_attr
            self._bi_in_id_dict = None if self._net is None else network._edge_in
            self._bi_out_id_dict = None if self._net is None else network._edge_out
            self._bi_id_attr = None if self._net is None else network._edge_attr
        elif self._id_kind == "diedge":
            self._in_id_dict = None if self._net is None else network._edge_in
            self._out_id_dict = None if self._net is None else network._edge_out
            self._id_attr = None if self._net is None else network._edge_attr
            self._bi_in_id_dict = None if self._net is None else network._node_in
            self._bi_out_id_dict = None if self._net is None else network._node_out
            self._bi_id_attr = None if self._net is None else network._node_attr

        if ids is None:
            self._ids = self._in_id_dict
        else:
            self._ids = ids

    def __getattr__(self, attr):
        stat = dispatch_stat(self._id_kind, self._net, self, attr)
        self.__dict__[attr] = stat
        return stat

    def multi(self, names):
        return dispatch_many_stats(self._id_kind, self._net, self, names)

    @property
    def ids(self):
        """The ids in this view.

        Notes
        -----
        Do not use this property for membership check. Instead of `x in view.ids`,
        always use `x in view`.  The latter is always faster.

        """
        return set(self._in_id_dict) if self._ids is None else self._ids

    def __len__(self):
        """The number of IDs."""
        return len(self._in_id_dict) if self._ids is None else len(self._ids)

    def __iter__(self):
        """Returns an iterator over the IDs."""
        if self._ids is None:
            return iter({}) if self._in_id_dict is None else iter(self._in_id_dict)
        else:
            return iter(self._ids)

    def __getitem__(self, id):
        """Get the attributes of the ID.

        Parameters
        ----------
        id : hashable
            node or edge ID

        Returns
        -------
        dict
            Node attributes.

        Raises
        ------
        XGIError
            If the id is not being kept track of by this view, or if id is not in the
            hypergraph, or if id is not hashable.

        """
        if id not in self:
            raise IDNotFound(f"The ID {id} is not in this view")
        return self._id_attr[id]

    def __contains__(self, id):
        """Checks whether the ID is in the dihypergraph"""
        return id in self._ids

    def __str__(self):
        """Returns a string of the list of IDs."""
        return str(list(self))

    def __repr__(self):
        """Returns a summary of the class"""
        return f"{self.__class__.__name__}({tuple(self)})"

    def __call__(self, bunch):
        """Filter to the given bunch.

        Parameters
        ----------
        bunch : Iterable
            Iterable of IDs

        Returns
        -------
        IDView
            A new view that keeps track only of the IDs in the bunch.

        """
        return self.from_view(self, bunch)

    def filterby(self, stat, val, mode="eq"):
        """Filter the IDs in this view by a statistic.

        Parameters
        ----------
        stat : str or :class:`xgi.stats.DiNodeStat`/:class:`xgi.stats.DiEdgeStat`
            `DiNodeStat`/`DiEdgeStat` object, or name of a `DiNodeStat`/`DiEdgeStat`.
        val : Any
            Value of the statistic.  Usually a single numeric value.  When mode is
            'between', must be a tuple of exactly two values.
        mode : str, optional
            How to compare each value to `val`.  Can be one of the following.

            * 'eq' (default): Return IDs whose value is exactly equal to `val`.
            * 'neq': Return IDs whose value is not equal to `val`.
            * 'lt': Return IDs whose value is less than `val`.
            * 'gt': Return IDs whose value is greater than `val`.
            * 'leq': Return IDs whose value is less than or equal to `val`.
            * 'geq': Return IDs whose value is greater than or equal to `val`.
            * 'between': In this mode, `val` must be a tuple `(val1, val2)`.  Return IDs
              whose value `v` satisfies `val1 <= v <= val2`.

        See Also
        --------
        IDView.filterby_attr : For more details, see the `tutorial
        <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

        Examples
        --------
        By default, return the IDs whose value of the statistic is exactly equal to
        `val`.

        >>> import xgi
        >>> H = xgi.DiHypergraph([([1, 2, 3], [2, 3, 4, 5]), ([3, 4, 5], [])])
        >>> n = H.nodes
        >>> n.filterby('degree', 2)
        DiNodeView((3, 4, 5))

        Can choose other comparison methods via `mode`.

        >>> n.filterby('degree', 2, 'eq')
        DiNodeView((3, 4, 5))
        >>> n.filterby('degree', 2, 'neq')
        DiNodeView((1, 2))
        >>> n.filterby('degree', 2, 'lt')
        DiNodeView((1, 2))
        >>> n.filterby('degree', 2, 'gt')
        DiNodeView(())
        >>> n.filterby('degree', 2, 'leq')
        DiNodeView((1, 2, 3, 4, 5))
        >>> n.filterby('degree', 2, 'geq')
        DiNodeView((3, 4, 5))
        >>> n.filterby('degree', (2, 3), 'between')
        DiNodeView((3, 4, 5))
        """
        if not isinstance(stat, IDStat):
            try:
                stat = getattr(self, stat)
            except AttributeError as e:
                raise AttributeError(f'Statistic with name "{stat}" not found') from e

        values = stat.asdict()
        if mode == "eq":
            bunch = [idx for idx in self if values[idx] == val]
        elif mode == "neq":
            bunch = [idx for idx in self if values[idx] != val]
        elif mode == "lt":
            bunch = [idx for idx in self if values[idx] < val]
        elif mode == "gt":
            bunch = [idx for idx in self if values[idx] > val]
        elif mode == "leq":
            bunch = [idx for idx in self if values[idx] <= val]
        elif mode == "geq":
            bunch = [idx for idx in self if values[idx] >= val]
        elif mode == "between":
            bunch = [node for node in self if val[0] <= values[node] <= val[1]]
        else:
            raise ValueError(
                f"Unrecognized mode {mode}. mode must be one of 'eq', 'neq', 'lt', 'gt', 'leq', 'geq', or 'between'."
            )
        return type(self).from_view(self, bunch)

    def filterby_attr(self, attr, val, mode="eq", missing=None):
        """Filter the IDs in this view by an attribute.

        Parameters
        ----------
        attr : string
            The name of the attribute
        val : Any
            A single value or, in the case of 'between', a list of length 2
        mode : str, optional
            Comparison mode. Valid options are 'eq' (default), 'neq', 'lt', 'gt',
            'leq', 'geq', or 'between'.
        missing : Any, optional
            The default value if the attribute is missing. If None (default),
            ignores those IDs.


        See Also
        --------
        DiIDView.filterby : Identical method.  For more details, see the `tutorial
        <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

        Notes
        -----
        Beware of using comparison modes ("lt", "gt", "leq", "geq")
        when the attribute is a string. For example, the string comparison
        `'10' < '9'` evaluates to `True`.
        """
        attrs = dispatch_stat(self._id_kind, self._net, self, "attrs")
        values = attrs(attr, missing).asdict()

        if mode == "eq":
            bunch = [
                idx for idx in self if values[idx] is not None and values[idx] == val
            ]
        elif mode == "neq":
            bunch = [
                idx for idx in self if values[idx] is not None and values[idx] != val
            ]
        elif mode == "lt":
            bunch = [
                idx for idx in self if values[idx] is not None and values[idx] < val
            ]
        elif mode == "gt":
            bunch = [
                idx for idx in self if values[idx] is not None and values[idx] > val
            ]
        elif mode == "leq":
            bunch = [
                idx for idx in self if values[idx] is not None and values[idx] <= val
            ]
        elif mode == "geq":
            bunch = [
                idx for idx in self if values[idx] is not None and values[idx] >= val
            ]
        elif mode == "between":
            bunch = [
                idx
                for idx in self
                if values[idx] is not None and val[0] <= values[idx] <= val[1]
            ]
        else:
            raise ValueError(
                f"Unrecognized mode {mode}. mode must be one of 'eq', 'neq', 'lt', 'gt', 'leq', 'geq', or 'between'."
            )
        return type(self).from_view(self, bunch)

    @classmethod
    def from_view(cls, view, bunch=None):
        """Create a view from another view.

        Allows to create a view with the same underlying data but with a different
        bunch.

        Parameters
        ----------
        view : IDView
            The view used to initialze the new object
        bunch : iterable
            IDs the new view will keep track of

        Returns
        -------
        DiIDView
            A view that is identical to `view` but keeps track of different IDs.

        """
        newview = cls(None)
        newview._net = view._net
        newview._in_id_dict = view._in_id_dict
        newview._out_id_dict = view._out_id_dict
        newview._id_attr = view._id_attr
        newview._bi_in_id_dict = view._bi_in_id_dict
        newview._bi_out_id_dict = view._bi_out_id_dict
        newview._bi_id_attr = view._bi_id_attr
        all_ids = set(view._in_id_dict)
        if bunch is None:
            newview._ids = all_ids
        else:
            bunch = set(bunch)
            wrong = bunch - all_ids
            if wrong:
                raise IDNotFound(f"IDs {wrong} not in the hypergraph")
            newview._ids = bunch
        return newview

    def _from_iterable(self, it):
        """Construct an instance of the class from any iterable input.

        This overrides collections.abc.Set._from_iterable, which is in turn used to
        implement set operations such as &, |, ^, -.

        """
        return self.from_view(self, it)


class DiNodeView(DiIDView):
    """An DiIDView that keeps track of node ids.

    .. warning::
        This is currently an experimental feature.

    Parameters
    ----------
    hypergraph : DiHypergraph
        The hypergraph whose nodes this view will keep track of.
    bunch : optional iterable, default None
        The node ids to keep track of.  If None (default), keep track of all node ids.

    See Also
    --------
    DiIDView

    Notes
    -----
    In addition to the methods listed in this page, other methods defined in the `stats`
    package are also accessible via the `NodeView` class.  For more details, see the
    `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """

    _id_kind = "dinode"

    def __init__(self, H, bunch=None):
        if H is None:
            super().__init__(None, bunch)
        else:
            super().__init__(H, bunch)

    def dimemberships(self, n=None):
        """Get the edge ids of which a node is a member.

        Gets all the node memberships for all nodes in the view if n
        not specified.

        Parameters
        ----------
        n : hashable, optional
            Node ID. By default, None.

        Returns
        -------
        dict of directed node memberships if n is None,
            otherwise the directed memberships of a single node.

        Raises
        ------
        XGIError
            If `n` is not hashable or if it is not in the hypergraph.

        """
        return (
            {
                key: (self._out_id_dict[key].copy(), self._in_id_dict[key].copy())
                for key in self
            }
            if n is None
            else (self._out_id_dict[n].copy(), self._in_id_dict[n].copy())
        )

    def memberships(self, n=None):
        """Get the edge ids of which a node is a member.

        Gets all the node memberships for all nodes in the view if n
        not specified.

        Parameters
        ----------
        n : hashable, optional
            Node ID. By default, None.

        Returns
        -------
        dict of sets if n is None, otherwise a set
            Node memberships, regardless of whether
            that node is a sender or receiver.

        Raises
        ------
        XGIError
            If `n` is not hashable or if it is not in the dihypergraph.

        """
        return (
            {
                key: set(self._out_id_dict[key].union(self._in_id_dict[key]))
                for key in self
            }
            if n is None
            else set(self._out_id_dict[n].union(self._in_id_dict[n]))
        )

    def isolates(self):
        """Nodes that belong to no edges.

        When ignore_singletons is True, a node is considered isolated from the rest of
        the hypergraph when it is included in no edges of size two or more.  In
        particular, whether the node is part of any singleton edges is irrelevant to
        determine whether it is isolated.

        When ignore_singletons is False (default), a node is isolated only when it is a
        member of exactly zero edges, including singletons.

        Returns
        -------
        NodeView containing the isolated nodes.

        See Also
        --------
        :meth:`EdgeView.singletons`

        """
        return self.filterby("degree", 0)


class DiEdgeView(DiIDView):
    """An DiIDView that keeps track of edge ids.

    .. warning::
        This is currently an experimental feature.

    Parameters
    ----------
    hypergraph : DiHypergraph
        The hypergraph whose edges this view will keep track of.
    bunch : optional iterable, default None
        The edge ids to keep track of.  If None (default), keep track of all edge ids.

    See Also
    --------
    DiIDView

    Notes
    -----
    In addition to the methods listed in this page, other methods defined in the `stats`
    package are also accessible via the `EdgeView` class.  For more details, see the
    `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    """

    _id_kind = "diedge"

    def __init__(self, H, bunch=None):
        if H is None:
            super().__init__(None, bunch)
        else:
            super().__init__(H, bunch)

    def dimembers(self, e=None, dtype=list):
        """Get the node ids that are members of an edge.

        Parameters
        ----------
        e : hashable, optional
            Edge ID. By default, None.
        dtype : {list, dict}, optional
            Specify the type of the return value.
            By default, list.

        Returns
        -------
        list (if dtype is list, default)
            Directed edges.
        dict (if dtype is dict)
            Directed edges.
        set (if e is not None)
            A single directed edge.

        In all of these cases, a directed edge is
        a 2-tuple of sets, where the first entry
        is the tail, and the second entry is the head.

        Raises
        ------
        TypeError
            If `e` is not None or a hashable
        XGIError
            If `dtype` is not dict or list
        IDNotFound
            If `e` does not exist in the hypergraph

        """
        if e is None:
            if dtype is dict:
                return {
                    key: (self._out_id_dict[key].copy(), self._in_id_dict[key].copy())
                    for key in self
                }
            elif dtype is list:
                return [
                    (self._out_id_dict[key].copy(), self._in_id_dict[key].copy())
                    for key in self
                ]
            else:
                raise XGIError(f"Unrecognized dtype {dtype}")

        if e not in self:
            raise IDNotFound(f'ID "{e}" not in this view')

        return (self._out_id_dict[e].copy(), self._in_id_dict[e].copy())

    def members(self, e=None, dtype=list):
        """Get the edges of a directed hypergraph.

        Parameters
        ----------
        e : hashable, optional
            Edge ID. By default, None.
        dtype : {list, dict}, optional
            Specify the type of the return value.
            By default, list.

        Returns
        -------
        list (if dtype is list, default)
            Edge members.
        dict (if dtype is dict)
            Edge members.
        set (if e is not None)
            Members of edge e.

        The members of an edge are the union of
        its head and tail sets.

        The

        Raises
        ------
        TypeError
            If `e` is not None or a hashable
        XGIError
            If `dtype` is not dict or list
        IDNotFound
            If `e` does not exist in the hypergraph

        """
        if e is None:
            if dtype is dict:
                return {
                    key: set(self._out_id_dict[key].union(self._in_id_dict[key]))
                    for key in self
                }
            elif dtype is list:
                return [
                    set(self._out_id_dict[key].union(self._in_id_dict[key]))
                    for key in self
                ]
            else:
                raise XGIError(f"Unrecognized dtype {dtype}")

        if e not in self:
            raise IDNotFound(f'ID "{e}" not in this view')

        return set(self._out_id_dict[e].union(self._in_id_dict[e]))

    def head(self, e=None, dtype=list):
        """Get the node ids that are in the head of a directed edge.

        Parameters
        ----------
        e : hashable, optional
            Edge ID. By default, None.
        dtype : {list, dict}, optional
            Specify the type of the return value.
            By default, list.

        Returns
        -------
        list (if dtype is list, default)
            Head members.
        dict (if dtype is dict)
            Head members.
        set (if e is not None)
            Members of the head of edge e.

        Raises
        ------
        TypeError
            If `e` is not None or a hashable
        XGIError
            If `dtype` is not dict or list
        IDNotFound
            If `e` does not exist in the hypergraph

        """
        if e is None:
            if dtype is dict:
                return {key: self._in_id_dict[key].copy() for key in self}
            elif dtype is list:
                return [self._in_id_dict[key].copy() for key in self]
            else:
                raise XGIError(f"Unrecognized dtype {dtype}")

        if e not in self:
            raise IDNotFound(f'ID "{e}" not in this view')

        return self._in_id_dict[e].copy()

    def tail(self, e=None, dtype=list):
        """Get the node ids that are in the tail of a directed edge.

        Parameters
        ----------
        e : hashable, optional
            Edge ID. By default, None.
        dtype : {list, dict}, optional
            Specify the type of the return value.
            By default, list.

        Returns
        -------
        list (if dtype is list, default)
            Tail members.
        dict (if dtype is dict)
            Tail members.
        set (if e is not None)
            Tail members of edge e.

        Raises
        ------
        TypeError
            If `e` is not None or a hashable
        XGIError
            If `dtype` is not dict or list
        IDNotFound
            If `e` does not exist in the hypergraph

        """
        if e is None:
            if dtype is dict:
                return {key: self._out_id_dict[key].copy() for key in self}
            elif dtype is list:
                return [self._out_id_dict[key].copy() for key in self]
            else:
                raise XGIError(f"Unrecognized dtype {dtype}")

        if e not in self:
            raise IDNotFound(f'ID "{e}" not in this view')

        return self._out_id_dict[e].copy()

    def sources(self, e=None, dtype=list):
        """Get the nodes that are sources (senders)
        in the directed edges.

        See Also
        --------
        tail: identical method
        """
        return self.tail(e=e, dtype=dtype)

    def targets(self, e=None, dtype=list):
        """Get the nodes that are sources (senders)
        in the directed edges.

        See Also
        --------
        head: identical method

        """
        return self.head(e=e, dtype=dtype)



"""Algorithms related to connected components of a hypergraph."""

from ..core.globalviews import subhypergraph
from ..exception import XGIError

__all__ = [
    "is_connected",
    "connected_components",
    "largest_connected_component",
    "number_connected_components",
    "node_connected_component",
    "largest_connected_hypergraph",
]


def is_connected(H):
    """
    A function to determine whether a hypergraph is connected.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest

    Returns
    -------
    bool
        Whether the hypergraph is connected.

    See Also
    --------
    connected_components
    number_connected_components
    largest_connected_component
    largest_connected_hypergraph

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(10, [0.5, 0.01], seed=1)
    >>> print(xgi.is_connected(H))
    True

    """
    return len(_plain_bfs(H, list(H.nodes)[0])) == len(H)


def connected_components(H):
    """
    A function to find the connected components of a hypergraph.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest

    Returns
    -------
    iterable of sets
        An iterator where each entry is a component of the hypergraph.

    See Also
    --------
    is_connected
    number_connected_components
    largest_connected_component
    largest_connected_hypergraph

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.1, 0.01], seed=1)
    >>> print([len(component) for component in xgi.connected_components(H)])
    [50]

    """
    seen = set()
    for v in H:
        if v not in seen:
            c = _plain_bfs(H, v)
            seen.update(c)
            yield c


def number_connected_components(H):
    """
    A function to find the number of connected components of a hypergraph.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest

    Returns
    -------
    int
        The number of connected components of the hypergraph.

    See Also
    --------
    is_connected
    connected_components
    largest_connected_component
    largest_connected_hypergraph

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.1, 0.01], seed=1)
    >>> print(xgi.number_connected_components(H))
    1

    """
    num_cc = 0
    seen = set()
    for v in H:
        if v not in seen:
            c = _plain_bfs(H, v)
            seen.update(c)
            num_cc += 1
    return num_cc


def largest_connected_component(H):
    """
    A function to find the largest connected component of a hypergraph.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest

    Returns
    -------
    set
        The largest connected component (a set of nodes) of the hypergraph.

    See Also
    --------
    connected_components
    largest_connected_hypergraph

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.1, 0.01], seed=1)
    >>> print(len(xgi.largest_connected_component(H)))
    50

    """
    return max(connected_components(H), key=len)


def node_connected_component(H, n):
    """
    A function to find the connected component of which a node in the
    hypergraph is a part.

    Parameters
    ----------
    H: Hypergraph object
        The hypergraph of interest
    n: hashable
        Node label

    See Also
    --------
    connected_components

    Returns
    -------
    set
        Returns the connected component of which the specified node in the
        hypergraph is a part.

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(50, [0.1, 0.01], seed=1)
    >>> comp = xgi.node_connected_component(H, 0)
    >>> print(type(comp), len(comp))
    <class 'set'> 50

    """
    if n in H:
        return _plain_bfs(H, n)
    else:
        raise XGIError("Specified node is not in the hypergraph!")


def _plain_bfs(H, source):
    """A fast BFS node generator"""
    seen = set()
    nextlevel = {source}
    while nextlevel:
        thislevel = nextlevel
        nextlevel = set()
        for v in thislevel:
            if v not in seen:
                seen.add(v)
                nextlevel.update(H.nodes.neighbors(v))
    return seen


def largest_connected_hypergraph(H, in_place=False):
    """
    A function to find the largest connected hypergraph from a data set.

    Parameters
    ----------
    H: Hypergraph
        The hypergraph of interest
    in_place: bool, optional
        If False, creates a copy; if True, modifies the existing hypergraph.
        By default, True.

    Returns
    -------
    None
        If in_place: modifies the existing hypergraph

    Hypergraph
        If not in_place: the hypergraph induced on the nodes of the
        largest connected component.

    See Also
    --------
    connected_components
    largest_connected_component

    Example
    -------
    >>> import xgi
    >>> H = xgi.random_hypergraph(10, [0.1, 0.01], seed=1)
    >>> H_gcc = xgi.largest_connected_hypergraph(H)
    >>> print(H_gcc.num_nodes)
    8

    """
    connected_nodes = max(connected_components(H), key=len)
    if not in_place:
        return subhypergraph(H, nodes=connected_nodes).copy()
    else:
        H.remove_nodes_from(set(H.nodes).difference(connected_nodes))



"""Base class for undirected hypergraphs."""
from collections import defaultdict
from collections.abc import Hashable, Iterable
from copy import copy, deepcopy
from itertools import count
from warnings import warn

from ..exception import IDNotFound, XGIError, frozen
from ..utils import IDDict, update_uid_counter
from .views import EdgeView, NodeView

__all__ = ["Hypergraph"]


class Hypergraph:
    r"""A hypergraph is a collection of subsets of a set of *nodes* or *vertices*.

    A hypergraph is a pair :math:`(V, E)`, where :math:`V` is a set of elements called
    *nodes* or *vertices*, and :math:`E` is a set whose elements are subsets of
    :math:`V`, that is, each :math:`e \in E` satisfies :math:`e \subset V`.  The
    elements of :math:`E` are called *hyperedges* or simply *edges*.

    The Hypergraph class allows any hashable object as a node and can associate
    attributes to each node, edge, or the hypergraph itself, in the form of key/value
    pairs. In this representation, multiedges are allowed.

    Parameters
    ----------
    incoming_data : input hypergraph data, optional
        Data to initialize the hypergraph. If None (default), an empty
        hypergraph is created, i.e. one with no nodes or edges.
        The data can be in the following formats:

        * hyperedge list
        * hyperedge dictionary
        * 2-column Pandas dataframe (bipartite edges)
        * Incidence matrix: numpy ndarray or scipy.sparse array
        * Hypergraph object
        * SimplicialComplex object

    **attr : dict, optional
        Attributes to add to the hypergraph as key, value pairs.
        By default, None.

    Notes
    -----
    Unique IDs are assigned to each node and edge internally and are used to refer to
    them throughout.

    The `attr` keyword arguments are added as hypergraph attributes. To add node or edge
    attributes see :meth:`add_node` and :meth:`add_edge`.

    In addition to the methods listed in this page, other methods defined in the `stats`
    package are also accessible via the `Hypergraph` class.  For more details, see the
    `tutorial
    <https://github.com/xgi-org/xgi/blob/main/tutorials/Tutorial%206%20-%20Statistics.ipynb>`_.

    Examples
    --------
    >>> import xgi
    >>> H = xgi.Hypergraph([[1, 2, 3], [4], [5, 6], [6, 7, 8]])
    >>> H.nodes
    NodeView((1, 2, 3, 4, 5, 6, 7, 8))
    >>> H.edges
    EdgeView((0, 1, 2, 3))

    """
    _node_dict_factory = IDDict
    _node_attr_dict_factory = IDDict
    _hyperedge_dict_factory = IDDict
    _hyperedge_attr_dict_factory = IDDict
    _hypergraph_attr_dict_factory = dict

    def __getstate__(self):
        """Function that allows pickling.

        Returns
        -------
        dict
            The keys label the hyeprgraph dict and the values
            are dictionarys from the Hypergraph class.

        Notes
        -----
        This allows the python multiprocessing module to be used.

        """
        return {
            "_edge_uid": self._edge_uid,
            "_hypergraph": self._hypergraph,
            "_node": self._node,
            "_node_attr": self._node_attr,
            "_edge": self._edge,
            "_edge_attr": self._edge_attr,
        }

    def __setstate__(self, state):
        """Function that allows unpickling of a hypergraph.

        Parameters
        ----------
        state
            The keys access the dictionary names the values are the
            dictionarys themselves from the Hypergraph class.

        Notes
        -----
        This allows the python multiprocessing module to be used.
        """
        self._edge_uid = state["_edge_uid"]
        self._hypergraph = state["_hypergraph"]
        self._node = state["_node"]
        self._node_attr = state["_node_attr"]
        self._edge = state["_edge"]
        self._edge_attr = state["_edge_attr"]
        self._nodeview = NodeView(self)
        self._edgeview = EdgeView(self)

    def __init__(self, incoming_data=None, **attr):
        self._edge_uid = count()
        self._hypergraph = self._hypergraph_attr_dict_factory()
        self._node = self._node_dict_factory()
        self._node_attr = self._node_attr_dict_factory()
        self._edge = self._hyperedge_dict_factory()
        self._edge_attr = self._hyperedge_attr_dict_factory()

        self._nodeview = NodeView(self)
        """A :class:`~xgi.core.reportviews.NodeView` of the hypergraph."""

        self._edgeview = EdgeView(self)
        """An :class:`~xgi.core.reportviews.EdgeView` of the hypergraph."""

        if incoming_data is not None:
            # This import needs to happen when this function is called, not when it is
            # defined.  Otherwise, a circular import error would happen.
            from ..convert import convert_to_hypergraph

            convert_to_hypergraph(incoming_data, create_using=self)
        self._hypergraph.update(attr)  # must be after convert

    def __str__(self):
        """Returns a short summary of the hypergraph.

        Returns
        -------
        string
            Hypergraph information

        """
        try:
            return (
                f"{type(self).__name__} named {self['name']} "
                f"with {self.num_nodes} nodes and {self.num_edges} hyperedges"
            )
        except XGIError:
            return (
                f"Unnamed {type(self).__name__} with "
                f"{self.num_nodes} nodes and {self.num_edges} hyperedges"
            )

    def __iter__(self):
        """Iterate over the nodes.

        Returns
        -------
        iterator
            An iterator over all nodes in the hypergraph.
        """
        return iter(self._node)

    def __contains__(self, n):
        """Check for if a node is in this hypergraph.

        Parameters
        ----------
        n : hashable
            node ID

        Returns
        -------
        bool
            Whether the node exists in the hypergraph.
        """
        try:
            return n in self._node
        except TypeError:
            return False

    def __len__(self):
        """Number of nodes in the hypergraph.

        Returns
        -------
        int
            The number of nodes in the hypergraph.

        See Also
        --------
        num_nodes : identical method
        num_edges : number of edges in the hypergraph

        """
        return len(self._node)

    def __getitem__(self, attr):
        """Read hypergraph attribute."""
        try:
            return self._hypergraph[attr]
        except KeyError:
            raise XGIError("This attribute has not been set.")

    def __setitem__(self, attr, val):
        """Write hypergraph attribute."""
        self._hypergraph[attr] = val

    def __getattr__(self, attr):
        stat = getattr(self.nodes, attr, None)
        word = "nodes"
        if stat is None:
            stat = getattr(self.edges, attr, None)
            word = "edges"
        if stat is None:
            word = None
            raise AttributeError(
                f"{attr} is not a method of Hypergraph or a "
                "recognized NodeStat or EdgeStat"
            )

        def func(node=None, *args, **kwargs):
            val = stat(*args, **kwargs).asdict()
            return val if node is None else val[node]

        func.__doc__ = f"""Equivalent to H.{word}.{attr}.asdict(). For accepted *args and
        **kwargs, see documentation of H.{word}.{attr}."""

        return func

    def __lshift__(self, H2):
        """Adds the edges of a hypergraph to the current hypergraph
        and updates the attributes.

        The node/edge attributes of the new hypergraph take precedence.

        Relabels all the edge IDs to preserve all the edges but
        keeps the node labels the same.

        Parameters
        ----------
        H2 : Hypergraph
            The hypergraph to update with.

        Returns
        -------
        Hypergraph
            The updated hypergraph

        Notes
        -----
        Addition is not quite commutative; the attributes of nodes and edges
        may be overwritten depending on whether they are first or second
        to be added. In addition, the edge IDs are assigned based on the order
        in which the edges are added, but does not functionally change the
        structure of the hypergraph.

        Examples
        --------

        >>> import xgi
        >>> H1 = xgi.Hypergraph([[1, 2], [2, 3]])
        >>> H2 = xgi.Hypergraph([[1, 3, 4]])
        >>> H = H1 << H2
        >>> H.edges.members()
        [{1, 2}, {2, 3}, {1, 3, 4}]
        """
        tempH = Hypergraph()

        tempH.add_nodes_from(zip(self._node.keys(), self._node_attr.values()))
        tempH.add_nodes_from(zip(H2._node.keys(), H2._node_attr.values()))

        tempH.add_edges_from(zip(self._edge.values(), self._edge_attr.values()))
        tempH.add_edges_from(zip(H2._edge.values(), H2._edge_attr.values()))

        tempH._hypergraph = deepcopy(self._hypergraph)
        tempH._hypergraph.update(deepcopy(H2._hypergraph))

        return tempH

    @property
    def nodes(self):
        """A :class:`NodeView` of this network."""
        return self._nodeview

    @property
    def edges(self):
        """An :class:`EdgeView` of this network."""
        return self._edgeview

    @property
    def num_nodes(self):
        """The number of nodes in the hypergraph.

        Returns
        -------
        int
            The number of nodes in the hypergraph.

        See Also
        --------
        num_edges : returns the number of edges in the hypergraph

        Examples
        --------
        >>> import xgi
        >>> hyperedge_list = [[1, 2], [2, 3, 4]]
        >>> H = xgi.Hypergraph(hyperedge_list)
        >>> H.num_nodes
        4

        """
        return len(self._node)

    @property
    def num_edges(self):
        """The number of edges in the hypergraph.

        Returns
        -------
        int
            The number of edges in the hypergraph.

        See Also
        --------
        num_nodes : returns the number of nodes in the hypergraph

        Examples
        --------
        >>> import xgi
        >>> hyperedge_list = [[1, 2], [2, 3, 4]]
        >>> H = xgi.Hypergraph(hyperedge_list)
        >>> H.num_edges
        2

        """
        return len(self._edge)

    def add_node(self, node, **attr):
        """Add one node with optional attributes.

        Parameters
        ----------
        node : node
            A node can be any hashable Python object except None.
        attr : keyword arguments, optional
            Set or change node attributes using key=value.

        See Also
        --------
        add_nodes_from
        set_node_attributes

        Notes
        -----
        If node is already in the hypergraph, its attributes are still updated.

        """
        if node not in self._node:
            self._node[node] = set()
            self._node_attr[node] = self._node_attr_dict_factory()
        self._node_attr[node].update(attr)

    def add_nodes_from(self, nodes_for_adding, **attr):
        """Add multiple nodes with optional attributes.

        Parameters
        ----------
        nodes_for_adding : iterable
            An iterable of nodes (list, dict, set, etc.).
            OR
            An iterable of (node, attribute dict) tuples.
            Node attributes are updated using the attribute dict.
        attr : keyword arguments, optional (default= no attributes)
            Update attributes for all nodes in nodes.
            Node attributes specified in nodes as a tuple take
            precedence over attributes specified via keyword arguments.

        See Also
        --------
        add_node
        set_node_attributes
        """
        for n in nodes_for_adding:
            try:
                newnode = n not in self._node
                newdict = attr
            except TypeError:
                n, ndict = n
                newnode = n not in self._node
                newdict = attr.copy()
                newdict.update(ndict)
            if newnode:
                self._node[n] = set()
                self._node_attr[n] = self._node_attr_dict_factory()
            self._node_attr[n].update(newdict)

    def remove_node(self, n, strong=False):
        """Remove a single node.

        The removal may be weak (default) or strong.  In weak removal, the node is
        removed from each of its containing edges.  If it is contained in any singleton
        edges, then these are also removed.  In strong removal, all edges containing the
        node are removed, regardless of size.

        Parameters
        ----------
        n : node
            A node in the hypergraph

        strong : bool, optional
            Whether to execute weak or strong removal. By default, False.

        Raises
        ------
        XGIError
           If n is not in the hypergraph.

        See Also
        --------
        remove_nodes_from

        """
        edge_neighbors = self._node[n]
        del self._node[n]
        del self._node_attr[n]

        if strong:
            for e in edge_neighbors:
                node_neighbors = self._edge[e]
                del self._edge[e]
                del self._edge_attr[e]
                for node in node_neighbors.difference({n}):
                    self._node[node].remove(e)
        else:  # weak removal
            for edge in edge_neighbors:
                self._edge[edge].remove(n)
                if not self._edge[edge]:
                    del self._edge[edge]
                    del self._edge_attr[edge]

    def remove_nodes_from(self, nodes):
        """Remove multiple nodes.

        Parameters
        ----------
        nodes : iterable
            An iterable of nodes.

        See Also
        --------
        remove_node

        """
        for n in nodes:
            if n not in self:
                warn(f"Node {n} not in hypergraph")
                continue
            self.remove_node(n)

    def set_node_attributes(self, values, name=None):
        """Sets node attributes from a given value or dictionary of values.

        Parameters
        ----------
        values : scalar value, dict-like
            What the node attribute should be set to.  If `values` is
            not a dictionary, then it is treated as a single attribute value
            that is then applied to every node in `H`.  This means that if
            you provide a mutable object, like a list, updates to that object
            will be reflected in the node attribute for every node.
            The attribute name will be `name`.

            If `values` is a dict or a dict of dict, it should be keyed
            by node to either an attribute value or a dict of attribute key/value
            pairs used to update the node's attributes.
        name : string, optional
            Name of the node attribute to set if values is a scalar, by default None.

        See Also
        --------
        set_edge_attributes
        add_node
        add_nodes_from

        Notes
        -----
        After computing some property of the nodes of a hypergraph, you may
        want to assign a node attribute to store the value of that property
        for each node.

        If you provide a list as the second argument, updates to the list
        will be reflected in the node attribute for each node.

        If you provide a dictionary of dictionaries as the second argument,
        the outer dictionary is assumed to be keyed by node to an inner
        dictionary of node attributes for that node.

        Note that if the dictionary contains nodes that are not in `G`, the
        values are silently ignored.

        """
        # Set node attributes based on type of `values`
        if name is not None:  # `values` must not be a dict of dict
            if isinstance(values, dict):  # `values` is a dict
                for n, v in values.items():
                    try:
                        self._node_attr[n][name] = v
                    except IDNotFound:
                        warn(f"Node {n} does not exist!")
            else:  # `values` is a constant
                for n in self:
                    self._node_attr[n][name] = values
        else:  # `values` must be dict of dict
            try:
                for n, d in values.items():
                    try:
                        self._node_attr[n].update(d)
                    except IDNotFound:
                        warn(f"Node {n} does not exist!")
            except (TypeError, ValueError, AttributeError):
                raise XGIError("Must pass a dictionary of dictionaries")

    def add_edge(self, members, id=None, **attr):
        """Add one edge with optional attributes.

        Parameters
        ----------
        members : Iterable
            An iterable of the ids of the nodes contained in the new edge.
        id : hashable, optional
            Id of the new edge. If None (default), a unique numeric ID will be created.
        **attr : dict, optional
            Attributes of the new edge.

        Raises
        -----
        XGIError
            If `members` is empty.

        See Also
        --------
        add_edges_from : Add a collection of edges.
        set_edge_attributes

        Examples
        --------

        Add edges with or without specifying an edge id.

        >>> import xgi
        >>> H = xgi.Hypergraph()
        >>> H.add_edge([1, 2, 3])
        >>> H.add_edge([3, 4], id='myedge')
        >>> H.edges
        EdgeView((0, 'myedge'))

        Access attributes using square brackets.  By default no attributes are created.

        >>> H.edges[0]
        {}
        >>> H.add_edge([1, 4], color='red', place='peru')
        >>> H.edges
        EdgeView((0, 'myedge', 1))
        >>> H.edges[1]
        {'color': 'red', 'place': 'peru'}

        """
        members = set(members)
        if not members:
            raise XGIError("Cannot add an empty edge")

        if id in self._edge.keys():  # check that uid is not present yet
            warn(f"uid {id} already exists, cannot add edge {members}")
            return

        uid = next(self._edge_uid) if id is None else id

        self._edge[uid] = set()
        for node in members:
            if node not in self._node:
                self._node[node] = set()
                self._node_attr[node] = self._node_attr_dict_factory()
            self._node[node].add(uid)
            self._edge[uid].add(node)

        self._edge_attr[uid] = self._hyperedge_attr_dict_factory()
        self._edge_attr[uid].update(attr)

        if id:  # set self._edge_uid correctly
            update_uid_counter(self, id)

    def add_edges_from(self, ebunch_to_add, **attr):
        r"""Add multiple edges with optional attributes.

        Parameters
        ----------
        ebunch_to_add : Iterable

            An iterable of edges.  This may be an iterable of iterables (Format 1),
            where each element contains the members of the edge specified as valid node
            IDs.  Alternatively, each element could also be a tuple in any of the
            following formats:

            * Format 2: 2-tuple (members, edge_id), or
            * Format 3: 2-tuple (members, attr), or
            * Format 4: 3-tuple (members, edge_id, attr),

            where `members` is an iterable of node IDs, `edge_id` is a hashable to use
            as edge ID, and `attr` is a dict of attributes. Finally, `ebunch_to_add` may
            be a dict of the form `{edge_id: edge_members}` (Format 5).

            Formats 2 and 3 are unambiguous because `attr` dicts are not hashable, while
            `id`s must be.  In Formats 2-4, each element of `ebunch_to_add` must have
            the same length, i.e. you cannot mix different formats.  The iterables
            containing edge members cannot be strings.

        attr : \*\*kwargs, optional
            Additional attributes to be assigned to all edges. Attribues specified via
            `ebunch_to_add` take precedence over `attr`.

        See Also
        --------
        add_edge : Add a single edge.
        add_weighted_edges_from : Convenient way to add weighted edges.
        set_edge_attributes

        Notes
        -----
        Adding the same edge twice will create a multi-edge. Currently
        cannot add empty edges; the method skips over them.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph()

        When specifying edges by their members only, numeric edge IDs will be assigned
        automatically.

        >>> H.add_edges_from([[0, 1], [1, 2], [2, 3, 4]])
        >>> H.edges.members(dtype=dict)
        {0: {0, 1}, 1: {1, 2}, 2: {2, 3, 4}}

        Custom edge ids can be specified using a dict.

        >>> H = xgi.Hypergraph()
        >>> H.add_edges_from({'one': [0, 1], 'two': [1, 2], 'three': [2, 3, 4]})
        >>> H.edges.members(dtype=dict)
        {'one': {0, 1}, 'two': {1, 2}, 'three': {2, 3, 4}}

        You can use the dict format to easily add edges from another hypergraph.

        >>> H2 = xgi.Hypergraph()
        >>> H2.add_edges_from(H.edges.members(dtype=dict))
        >>> H.edges == H2.edges
        True

        Alternatively, edge ids can be specified using an iterable of 2-tuples.

        >>> H = xgi.Hypergraph()
        >>> H.add_edges_from([([0, 1], 'one'), ([1, 2], 'two'), ([2, 3, 4], 'three')])
        >>> H.edges.members(dtype=dict)
        {'one': {0, 1}, 'two': {1, 2}, 'three': {2, 3, 4}}

        Attributes for each edge may be specified using a 2-tuple for each edge.
        Numeric IDs will be assigned automatically.

        >>> H = xgi.Hypergraph()
        >>> edges = [
        ...     ([0, 1], {'color': 'red'}),
        ...     ([1, 2], {'age': 30}),
        ...     ([2, 3, 4], {'color': 'blue', 'age': 40}),
        ... ]
        >>> H.add_edges_from(edges)
        >>> {e: H.edges[e] for e in H.edges}
        {0: {'color': 'red'}, 1: {'age': 30}, 2: {'color': 'blue', 'age': 40}}

        Attributes and custom IDs may be specified using a 3-tuple for each edge.

        >>> H = xgi.Hypergraph()
        >>> edges = [
        ...     ([0, 1], 'one', {'color': 'red'}),
        ...     ([1, 2], 'two', {'age': 30}),
        ...     ([2, 3, 4], 'three', {'color': 'blue', 'age': 40}),
        ... ]
        >>> H.add_edges_from(edges)
        >>> {e: H.edges[e] for e in H.edges}
        {'one': {'color': 'red'}, 'two': {'age': 30}, 'three': {'color': 'blue', 'age': 40}}

        """
        # format 5 is the easiest one
        if isinstance(ebunch_to_add, dict):
            for id, members in ebunch_to_add.items():
                if id in self._edge.keys():  # check that uid is not present yet
                    warn(f"uid {id} already exists, cannot add edge {members}.")
                    continue
                try:
                    self._edge[id] = set(members)
                except TypeError as e:
                    raise XGIError("Invalid ebunch format") from e
                for n in members:
                    if n not in self._node:
                        self._node[n] = set()
                        self._node_attr[n] = self._node_attr_dict_factory()
                    self._node[n].add(id)
                self._edge_attr[id] = self._hyperedge_attr_dict_factory()

                update_uid_counter(self, id)

            return

        # in formats 1-4 we only know that ebunch_to_add is an iterable, so we iterate
        # over it and use the firs element to determine which format we are working with
        new_edges = iter(ebunch_to_add)
        try:
            first_edge = next(new_edges)
        except StopIteration:
            return
        try:
            first_elem = list(first_edge)[0]
        except TypeError:
            first_elem = None

        format1, format2, format3, format4 = False, False, False, False
        if isinstance(first_elem, Iterable):
            if all(isinstance(e, str) for e in first_edge):
                format1 = True
            elif len(first_edge) == 2 and issubclass(type(first_edge[1]), Hashable):
                format2 = True
            elif len(first_edge) == 2:
                format3 = True
            elif len(first_edge) == 3:
                format4 = True
        else:
            format1 = True

        if (format1 and isinstance(first_edge, str)) or (
            not format1 and isinstance(first_elem, str)
        ):
            raise XGIError("Members cannot be specified as a string")

        # now we may iterate over the rest
        e = first_edge
        while True:
            if format1:
                members, id, eattr = e, next(self._edge_uid), {}
            elif format2:
                members, id, eattr = e[0], e[1], {}
            elif format3:
                members, id, eattr = e[0], next(self._edge_uid), e[1]
            elif format4:
                members, id, eattr = e[0], e[1], e[2]

            if id in self._edge.keys():  # check that uid is not present yet
                warn(f"uid {id} already exists, cannot add edge {members}.")
            else:
                try:
                    self._edge[id] = set(members)
                except TypeError as e:
                    raise XGIError("Invalid ebunch format") from e

                for n in members:
                    if n not in self._node:
                        self._node[n] = set()
                        self._node_attr[n] = self._node_attr_dict_factory()
                    self._node[n].add(id)

                self._edge_attr[id] = self._hyperedge_attr_dict_factory()
                self._edge_attr[id].update(attr)
                self._edge_attr[id].update(eattr)

            try:
                e = next(new_edges)
            except StopIteration:
                if format2 or format4:
                    update_uid_counter(self, id)
                break

    def add_weighted_edges_from(self, ebunch, weight="weight", **attr):
        """Add multiple weighted edges with optional attributes.

        Parameters
        ----------
        ebunch_to_add : iterable of edges
            Each edge given in the list or container will be added
            to the graph. The edges must be given as tuples of
            the form (node1, node2, ..., noden, weight).
        weight : string, optional
            The attribute name for the edge weights to be added,
            by default "weight".
        attr : keyword arguments, optional
            Edge attributes to add/update for all edges.

        See Also
        --------
        add_edge : Add a single edge.
        add_edges_from : Add multiple edges.
        set_edge_attributes
        get_edge_attributes

        Notes
        -----
        Adding the same edge twice creates a multiedge.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph()
        >>> edges = [(0, 1, 0.3), (0, 2, 0.8)]
        >>> H.add_weighted_edges_from(edges)
        >>> H.edges[0]
        {'weight': 0.3}

        """
        try:
            self.add_edges_from(
                ((edge[:-1], {weight: edge[-1]}) for edge in ebunch), **attr
            )
        except KeyError:
            XGIError("Empty or invalid edges specified.")

    def set_edge_attributes(self, values, name=None):
        """Set the edge attributes from a value or a dictionary of values.

        Parameters
        ----------
        values : scalar value, dict-like
            What the edge attribute should be set to.  If `values` is
            not a dictionary, then it is treated as a single attribute value
            that is then applied to every edge in `H`.  This means that if
            you provide a mutable object, like a list, updates to that object
            will be reflected in the edge attribute for each edge.  The attribute
            name will be `name`.
            If `values` is a dict or a dict of dict, it should be keyed
            by edge ID to either an attribute value or a dict of attribute
            key/value pairs used to update the edge's attributes.
        name : string, optional
            Name of the edge attribute to set if values is a scalar. By default, None.

        See Also
        --------
        set_node_attributes
        add_edge
        add_edges_from

        Notes
        -----
        Note that if the dict contains edge IDs that are not in `H`, they are
        silently ignored.

        """
        if name is not None:
            # `values` does not contain attribute names
            try:
                for e, value in values.items():
                    try:
                        self._edge_attr[id][name] = value
                    except IDNotFound:
                        warn(f"Edge {e} does not exist!")
            except AttributeError:
                # treat `values` as a constant
                for e in self._edge:
                    self._edge_attr[e][name] = values
        else:
            try:
                for e, d in values.items():
                    try:
                        self._edge_attr[e].update(d)
                    except IDNotFound:
                        warn(f"Edge {e} does not exist!")
            except AttributeError:
                raise XGIError(
                    "name property has not been set and a "
                    "dict-of-dicts has not been provided."
                )

    def double_edge_swap(self, n_id1, n_id2, e_id1, e_id2):
        """Swap the edge memberships of two selected nodes, given two edges.

        Parameters
        ----------
        n_id1 : hashable
            The ID of the first node, originally a member of the first edge.
        n_id2 : hashable
            The ID of the second node, originally a member of the second edge.
        e_id1 : hashable
            The ID of the first edge.
        e_id2 : hashable
            The ID of the second edge.

        Raises
        ------
        IDNotFound
            If user specifies nodes or edges that do not exist or
            nodes that are not part of edges.
        XGIError
            If the swap does not preserve edge sizes.

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph([[1, 2, 3], [3, 4]])
        >>> H.double_edge_swap(1, 4, 0, 1)
        >>> H.edges.members()
        [{2, 3, 4}, {1, 3}]

        """
        # Assign edges to modify
        try:
            # Initialize temporary copies to modify
            temp_memberships1 = self._node[n_id1].copy()
            temp_memberships2 = self._node[n_id2].copy()
            temp_members1 = self._edge[e_id1].copy()
            temp_members2 = self._edge[e_id2].copy()

            # remove old nodes from edges
            temp_members1.remove(n_id1)
            temp_members2.remove(n_id2)

            # swap nodes
            temp_members1.add(n_id2)
            temp_members2.add(n_id1)

            # Now we handle the memberships
            # remove old nodes from edges
            temp_memberships1.remove(e_id1)
            temp_memberships2.remove(e_id2)

            # swap nodes
            temp_memberships1.add(e_id2)
            temp_memberships2.add(e_id1)

        except KeyError as e:

            raise IDNotFound(
                "One of the nodes specified doesn't belong to the specified edge."
            ) from e

        if (
            len(temp_memberships1) != len(self._node[n_id1])
            or len(temp_memberships2) != len(self._node[n_id2])
            or len(temp_members1) != len(self._edge[e_id1])
            or len(temp_members2) != len(self._edge[e_id2])
        ):
            raise XGIError("This swap does not preserve edge sizes.")

        self._node[n_id1] = temp_memberships1
        self._node[n_id2] = temp_memberships2

        self._edge[e_id1] = temp_members1
        self._edge[e_id2] = temp_members2

    def add_node_to_edge(self, edge, node):
        """Add one node to an existing edge.

        If the node or edge IDs do not exist, they are created.

        Parameters
        ----------
        edge : hashable
            edge ID
        node : hashable
            node ID

        See Also
        --------
        add_node
        add_edge
        remove_node_from_edge

        Examples
        --------
        >>> import xgi
        >>> H = xgi.Hypergraph()
        >>> H.add_edge(['apple', 'banana'], 'fruits')
        >>> H.add_node_to_edge('fruits', 'pear')
        >>> H.add_node_to_edge('veggies', 'lettuce')
        >>> d = H.edges.members(dtype=dict)
        >>> {id: sorted(list(e)) for id, e in d.items()}
        {'fruits': ['apple', 'banana', 'pear'], 'veggies': ['lettuce']}

        """
        if edge not in self._edge:
            self._edge[edge] = set()
            self._edge_attr[edge] = {}
        if node not in self._node:
            self._node[node] = set()
            self._node_attr[node] = {}
        self._edge[edge].add(node)
        self._node[node].add(edge)

    def remove_edge(self, id):
        """Remove one edge.

        Parameters
        ----------
        id : Hashable
            edge ID to remove

        Raises
        ------
        XGIError
            If no edge has that ID.

        See Also
        --------
        remove_edges_from : Remove multiple edges.

        """
        for node in self._edge[id].copy():
            self._node[node].remove(id)
        del self._edge[id]
        del self._edge_attr[id]

    def remove_edges_from(self, ebunch):
        """Remove multiple edges.

        Parameters
        ----------
        ebunch: Iterable
            Edges to remove.

        Raises
        ------
        xgi.exception.IDNotFound
            If an id in ebunch is not part of the network.

        See Also
        --------
        remove_edge : remove a single edge.

        """
        for id in ebunch:
            for node in self._edge[id].copy():
                self._node[node].remove(id)
            del self._edge[id]
            del self._edge_attr[id]

    def remove_node_from_edge(self, edge, node):
        """Remove a node from an existing edge.

        Parameters
        ----------
        edge : hashable
            The edge ID
        node : hashable
            The node ID

        Raises
        ------
        XGIError
            If either the node or edge does not exist.

        See Also
        --------
        remove_node
        remove_edge
        add_node_to_edge

        Notes
        -----
        If edge is left empty as a result of removing node from it, the edge is also
        removed.

        """
        try:
            self._node[node].remove(edge)
        except KeyError as e:
            raise XGIError(f"Node {node} not in the hypergraph") from e
        except ValueError as e:
            raise XGIError(f"Node {node} not in edge {edge}") from e

        try:
            self._edge[edge].remove(node)
        except KeyError as e:
            raise XGIError(f"Edge {edge} not in the hypergraph") from e
        except ValueError as e:
            raise XGIError(f"Edge {edge} does not contain node {node}") from e

        if not self._edge[edge]:
            del self._edge[edge]
            del self._edge_attr[edge]

    def update(self, *, edges=None, nodes=None):
        """Add nodes or edges to the hypergraph.

        Parameters
        ----------
        edges : Iterable, optional
            Edges to be added. By default, None.
        nodes : Iterable, optional
            Nodes to be added. By default, None.

        See Also
        --------
        add_edges_from: Add multiple edges.
        add_nodes_from: Add multiple nodes.

        """
        if nodes:
            self.add_nodes_from(nodes)
        if edges:
            self.add_edges_from(edges)

    def clear(self, hypergraph_attr=True):
        """Remove all nodes and edges from the graph.

        Also removes node and edge attributes, and optionally hypergraph attributes.

        Parameters
        ----------
        hypergraph_attr : bool, optional
            Whether to remove hypergraph attributes as well.
            By default, True.

        """
        self._node.clear()
        self._node_attr.clear()
        self._edge.clear()
        self._edge_attr.clear()
        if hypergraph_attr:
            self._hypergraph.clear()

    def clear_edges(self):
        """Remove all edges from the graph without altering any nodes."""
        for node in self.nodes:
            self._node[node] = set()
        self._edge.clear()
        self._edge_attr.clear()

    def merge_duplicate_edges(
        self, rename="first", merge_rule="first", multiplicity=None
    ):
        """Merges edges which have the same members.

        Parameters
        ----------
        rename : str, optional
            Either "first" (default), "tuple", or "new".
            If "first", the new edge ID is the first of the sorted
            duplicate edge IDs. If "tuple", the new edge ID is a
            tuple of the sorted duplicate edge IDs. If "new", a
            new ID will be selected automatically.
        merge_rule : str, optional
            Either "first" (default) or "union".
            If "first", takes the attributes of the first duplicate.
            If "union", takes the set of attributes of all the duplicates.
        multiplicity : str, optional
            The attribute in which to store the multiplicity of the hyperedge,
            by default None.

        Raises
        ------
        XGIError
            If invalid rename or merge_rule specified.

        Warns
        -----
        If the user chooses merge_rule="union". Tells the
        user that they can no longer draw based on this stat.

        Examples
        --------

        >>> import xgi
        >>> edges = [{1, 2}, {1, 2}, {1, 2}, {3, 4, 5}, {3, 4, 5}]
        >>> edge_attrs = dict()
        >>> edge_attrs[0] = {"color": "blue"}
        >>> edge_attrs[1] = {"color": "red", "weight": 2}
        >>> edge_attrs[2] = {"color": "yellow"}
        >>> edge_attrs[3] = {"color": "purple"}
        >>> edge_attrs[4] = {"color": "purple", "name": "test"}
        >>> H = xgi.Hypergraph(edges)
        >>> H.set_edge_attributes(edge_attrs)
        >>> H.edges
        EdgeView((0, 1, 2, 3, 4))

        There are several ways to rename the duplicate edges after merging:

        1. The merged edge ID is the first duplicate edge ID.

        >>> H1 = H.copy()
        >>> H1.merge_duplicate_edges()
        >>> H1.edges
        EdgeView((0, 3))

        2. The merged edge ID is a tuple of all the duplicate edge IDs.

        >>> H2 = H.copy()
        >>> H2.merge_duplicate_edges(rename="tuple")
        >>> H2.edges
        EdgeView(((0, 1, 2), (3, 4)))

        3. The merged edge ID is assigned a new edge ID.

        >>> H3 = H.copy()
        >>> H3.merge_duplicate_edges(rename="new")
        >>> H3.edges
        EdgeView((5, 6))

        We can also specify how we would like to combine the attributes
        of the merged edges:

        1. The attributes are the attributes of the first merged edge.

        >>> H4 = H.copy()
        >>> H4.merge_duplicate_edges()
        >>> H4.edges[0]
        {'color': 'blue'}

        2. The attributes are the union of every attribute that each merged
        edge has. If a duplicate edge doesn't have that attribute, it is set
        to None.

        >>> H5 = H.copy()
        >>> H5.merge_duplicate_edges(merge_rule="union")
        >>> H5.edges[0] == {'color': {'blue', 'red', 'yellow'}, 'weight':{2, None}}
        True

        3. We can also set the attributes to the intersection, i.e.,
        if a particular attribute is the same across the duplicate
        edges, we use this attribute, otherwise, we set it to None.

        >>> H6 = H.copy()
        >>> H6.merge_duplicate_edges(merge_rule="intersection")
        >>> H6.edges[0] == {'color': None, 'weight': None}
        True
        >>> H6.edges[3] == {'color': 'purple', 'name': None}
        True

        We can also choose to store the multiplicity of the edge
        as an attribute. The user simply provides the string of
        the attribute which stores it. Note that this will not prevent
        other attributes from being over written (e.g., weight), so
        be careful that the attribute is not already in use.

        >>> H7 = H.copy()
        >>> H7.merge_duplicate_edges(multiplicity="mult")
        >>> H7.edges[0]['mult'] == 3
        True
        """
        dups = []
        hashes = defaultdict(list)
        for idx, members in self._edge.items():
            hashes[frozenset(members)].append(idx)

        new_edges = list()
        for members, dup_ids in hashes.items():
            if len(dup_ids) > 1:
                dups.extend(dup_ids)

                if rename == "first":
                    new_id = sorted(dup_ids)[0]
                elif rename == "tuple":
                    new_id = tuple(sorted(dup_ids))
                elif rename == "new":
                    new_id = next(self._edge_uid)
                else:
                    raise XGIError("Invalid ID renaming scheme!")

                if merge_rule == "first":
                    id = min(dup_ids)
                    new_attrs = deepcopy(self._edge_attr[id])
                elif merge_rule == "union":
                    attrs = {field for id in dup_ids for field in self._edge_attr[id]}
                    new_attrs = {
                        attr: {self._edge_attr[id].get(attr) for id in dup_ids}
                        for attr in attrs
                    }
                elif merge_rule == "intersection":
                    attrs = {field for id in dup_ids for field in self._edge_attr[id]}
                    set_attrs = {
                        attr: {self._edge_attr[id].get(attr) for id in dup_ids}
                        for attr in attrs
                    }
                    new_attrs = {
                        attr: (None if len(val) != 1 else next(iter(val)))
                        for attr, val in set_attrs.items()
                    }
                else:
                    raise XGIError("Invalid merge rule!")

                if multiplicity is not None:
                    new_attrs[multiplicity] = len(dup_ids)
                new_edges.append((members, new_id, new_attrs))
        self.remove_edges_from(dups)
        self.add_edges_from(new_edges)

        if merge_rule == "union":
            warn(
                "You will not be able to color/draw by "
                "merged attributes with xgi.draw()!"
            )

    def copy(self):
        """A deep copy of the hypergraph.

        A deep copy of the hypergraph, including node, edge, and hypergraph attributes.

        Returns
        -------
        H : Hypergraph
            A copy of the hypergraph.

        """
        cp = self.__class__()
        nn = self.nodes
        cp.add_nodes_from((n, deepcopy(attr)) for n, attr in nn.items())
        ee = self.edges
        cp.add_edges_from(
            (e, id, deepcopy(self.edges[id]))
            for id, e in ee.members(dtype=dict).items()
        )
        cp._hypergraph = deepcopy(self._hypergraph)

        cp._edge_uid = copy(self._edge_uid)

        return cp

    def dual(self):
        """The dual of the hypergraph.

        In the dual, nodes become edges and edges become nodes.

        Returns
        -------
        Hypergraph
            The dual of the hypergraph.

        """
        dual = self.__class__()
        nn = self.nodes
        dual.add_edges_from(
            (nn.memberships(n), n, deepcopy(attr)) for n, attr in nn.items()
        )
        ee = self.edges
        dual.add_nodes_from((e, deepcopy(attr)) for e, attr in ee.items())
        dual._hypergraph = deepcopy(self._hypergraph)

        return dual

    def cleanup(
        self,
        isolates=False,
        singletons=False,
        multiedges=False,
        connected=True,
        relabel=True,
        in_place=True,
    ):
        """Removes potentially undesirable artifacts from the hypergraph.

        Parameters
        ----------
        isolates : bool, optional
            Whether isolated nodes are allowed, by default False.
        singletons : bool, optional
            Whether singleton edges are allowed, by default False.
        multiedges : bool, optional
            Whether multiedges are allowed, by default False.
        connected : bool, optional
            Whether the returned hypergraph should be connected. If true,
            returns the hypergraph induced on the largest connected component.
            By default, False.
        relabel : bool, optional
            Whether to convert all node and edge labels to sequential integers, by
            default True.
        in_place : bool, optional
            Whether to modify the current hypergraph or output a new one, by default
            True.

        """
        if in_place:
            if not multiedges:
                self.merge_duplicate_edges()
            if not singletons:
                self.remove_edges_from(self.edges.singletons())
            if not isolates:
                self.remove_nodes_from(self.nodes.isolates())
            if connected:
                from ..algorithms import largest_connected_component

                self.remove_nodes_from(self.nodes - largest_connected_component(self))
            if relabel:
                from ..utils import convert_labels_to_integers

                temp = convert_labels_to_integers(self).copy()

                nn = temp.nodes
                ee = temp.edges

                self.clear()
                self.add_nodes_from((n, deepcopy(attr)) for n, attr in nn.items())
                self.add_edges_from(
                    (e, id, deepcopy(temp.edges[id]))
                    for id, e in ee.members(dtype=dict).items()
                )
                self._hypergraph = deepcopy(temp._hypergraph)
        else:
            H = self.copy()
            if not multiedges:
                H.merge_duplicate_edges()
            if not singletons:
                H.remove_edges_from(H.edges.singletons())
            if not isolates:
                H.remove_nodes_from(H.nodes.isolates())
            if connected:
                from ..algorithms import largest_connected_component

                H.remove_nodes_from(H.nodes - largest_connected_component(H))
            if relabel:
                from ..utils import convert_labels_to_integers

                H = convert_labels_to_integers(H)

            return H

    def freeze(self):
        """Method for freezing a hypergraph which prevents it from being modified

        See Also
        --------
        frozen : Method that raises an error when a user tries to modify the hypergraph
        is_frozen : Check whether a hypergraph is frozen

        Examples
        --------
        >>> import xgi
        >>> edges = [[1, 2], [2, 3, 4]]
        >>> H = xgi.Hypergraph(edges)
        >>> H.freeze()
        >>> H.add_node(5)
        Traceback (most recent call last):
        xgi.exception.XGIError: Frozen higher-order network can't be modified

        """
        self.add_node = frozen
        self.add_nodes_from = frozen
        self.remove_node = frozen
        self.remove_nodes_from = frozen
        self.add_edge = frozen
        self.add_edges_from = frozen
        self.add_weighted_edges_from = frozen
        self.remove_edge = frozen
        self.remove_edges_from = frozen
        self.add_node_to_edge = frozen
        self.remove_node_from_edge = frozen
        self.clear = frozen
        self.frozen = True

    @property
    def is_frozen(self):
        """Checks whether a dihypergraph is frozen

        Returns
        -------
        bool
            True if hypergraph is frozen, false if not.

        See Also
        --------
        freeze : A method to prevent a hypergraph from being modified.

        Examples
        --------
        >>> import xgi
        >>> edges = [[1, 2], [2, 3, 4]]
        >>> H = xgi.Hypergraph(edges)
        >>> H.freeze()
        >>> H.is_frozen
        True

        """
        try:
            return self.frozen
        except AttributeError:
            return False



from . import bigg_data, bipartite, edgelist, incidence, json, xgi_data
from .bipartite import *
from .edgelist import *
from .incidence import *
from .json import *
from .xgi_data import *
from .bigg_data import *



"""Read from and write to JSON."""
import json

from ..convert import dict_to_hypergraph
from ..exception import XGIError

__all__ = ["write_json", "read_json"]


def write_json(H, path):
    """
    A function to write a file in a standardized JSON format.

    Parameters
    ----------
    H: Hypergraph object
        The specified hypergraph object
    path: string
        The path of the file to read from

    """
    # initialize empty data
    data = {}

    # name always gets written (default is an empty string)
    data["hypergraph-data"] = {}
    data["hypergraph-data"].update(H._hypergraph)

    # get node data
    try:
        data["node-data"] = {str(idx): H.nodes[idx] for idx in H.nodes}
    except KeyError:
        raise XGIError("Node attributes not saved!")

    try:
        data["edge-data"] = {str(idx): H.edges[idx] for idx in H.edges}
    except KeyError:
        raise XGIError("Edge attributes not saved!")

    # hyperedge dict
    data["edge-dict"] = {
        str(idx): [str(n) for n in H.edges.members(idx)] for idx in H.edges
    }

    datastring = json.dumps(data, indent=2)

    with open(path, "w") as output_file:
        output_file.write(datastring)


def read_json(path, nodetype=None, edgetype=None):
    """
    A function to read a file in a standardized JSON format.

    Parameters
    ----------
    data: dict
        A dictionary in the hypergraph JSON format
    nodetype: type, optional
        type that the node IDs will be cast to
    edgetype: type, optional
        type that the edge IDs will be cast to

    Returns
    -------
    A Hypergraph object
        The loaded hypergraph

    Raises
    ------
    XGIError
        If the JSON is not in a format that can be loaded.

    """
    with open(path) as file:
        data = json.loads(file.read())

    return dict_to_hypergraph(data, nodetype=nodetype, edgetype=edgetype)



"""Methods for converting to and from bipartite graphs."""

import networkx as nx
from networkx import bipartite

from ..exception import XGIError
from ..generators import empty_hypergraph

__all__ = ["from_bipartite_graph", "to_bipartite_graph"]


def from_bipartite_graph(G, create_using=None, dual=False):
    """
    Create a Hypergraph from a NetworkX bipartite graph.

    Any hypergraph may be represented as a bipartite graph where
    nodes in the first layer are nodes and nodes in the second layer
    are hyperedges.

    The default behavior is to create nodes in the hypergraph
    from the nodes in the bipartite graph where the attribute
    bipartite=0 and hyperedges in the hypergraph from the nodes
    in the bipartite graph with attribute bipartite=1. Setting the
    keyword `dual` reverses this behavior.


    Parameters
    ----------
    G : nx.Graph
        A networkx bipartite graph. Each node in the graph has a property
        'bipartite' taking the value of 0 or 1 indicating the type of node.

    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None

    dual : bool, default : False
        If True, get edges from bipartite=0 and nodes from bipartite=1

    Returns
    -------
    Hypergraph
        The equivalent hypergraph

    References
    ----------
    The Why, How, and When of Representations for Complex Systems,
    Leo Torres, Ann S. Blevins, Danielle Bassett, and Tina Eliassi-Rad,
    https://doi.org/10.1137/20M1355896

    Examples
    --------
    >>> import networkx as nx
    >>> import xgi
    >>> G = nx.Graph()
    >>> G.add_nodes_from([1, 2, 3, 4], bipartite=0)
    >>> G.add_nodes_from(['a', 'b', 'c'], bipartite=1)
    >>> G.add_edges_from([(1, 'a'), (1, 'b'), (2, 'b'), (2, 'c'), (3, 'c'), (4, 'a')])
    >>> H = xgi.from_bipartite_graph(G)

    """
    edges = []
    nodes = []
    for n, d in G.nodes(data=True):
        try:
            node_type = d["bipartite"]
        except KeyError as e:
            raise XGIError("bipartite property not set") from e

        if node_type == 0:
            nodes.append(n)
        elif node_type == 1:
            edges.append(n)
        else:
            raise XGIError("Invalid type specifier")

    if not bipartite.is_bipartite_node_set(G, nodes):
        raise XGIError("The network is not bipartite")

    H = empty_hypergraph(create_using)
    H.add_nodes_from(nodes)
    for edge in edges:
        nodes_in_edge = list(G.neighbors(edge))
        H.add_edge(nodes_in_edge, id=edge)
    return H.dual() if dual else H


def to_bipartite_graph(H, index=False):
    """Create a NetworkX bipartite network from a hypergraph.

    Parameters
    ----------
    H: xgi.Hypergraph
        The XGI hypergraph object of interest
    index: bool (default False)
        If False (default), return only the graph.  If True, additionally return the
        index-to-node and index-to-edge mappings.

    Returns
    -------
    nx.Graph[, dict, dict]
        The resulting equivalent bipartite graph, and optionally the index-to-unit
        mappings.

    References
    ----------
    The Why, How, and When of Representations for Complex Systems,
    Leo Torres, Ann S. Blevins, Danielle Bassett, and Tina Eliassi-Rad,
    https://doi.org/10.1137/20M1355896

    Examples
    --------
    >>> import xgi
    >>> hyperedge_list = [[1, 2], [2, 3, 4]]
    >>> H = xgi.Hypergraph(hyperedge_list)
    >>> G = xgi.to_bipartite_graph(H)
    >>> G, itn, ite = xgi.to_bipartite_graph(H, index=True)

    """
    G = nx.Graph()

    n = H.num_nodes
    m = H.num_edges

    node_dict = dict(zip(H.nodes, range(n)))
    edge_dict = dict(zip(H.edges, range(n, n + m)))
    G.add_nodes_from(node_dict.values(), bipartite=0)
    G.add_nodes_from(edge_dict.values(), bipartite=1)
    for node in H.nodes:
        for edge in H.nodes.memberships(node):
            G.add_edge(node_dict[node], edge_dict[edge])

    if index:
        return (
            G,
            dict(zip(range(n), H.nodes)),
            dict(zip(range(n, n + m), H.edges)),
        )
    else:
        return G



"""Method for converting from a standardized dictionary."""

from ..exception import XGIError
from ..generators import empty_hypergraph

__all__ = ["dict_to_hypergraph"]


def dict_to_hypergraph(data, nodetype=None, edgetype=None, max_order=None):
    """
    A function to read a file in a standardized JSON format.

    Parameters
    ----------
    data: dict
        A dictionary in the hypergraph JSON format
    nodetype: type, optional
        Type that the node IDs will be cast to
    edgetype: type, optional
        Type that the edge IDs will be cast to
    max_order: int, optional
        Maximum order of edges to add to the hypergraph

    Returns
    -------
    A Hypergraph object
        The loaded hypergraph

    Raises
    ------
    XGIError
        If the JSON is not in a format that can be loaded.

    See Also
    --------
    read_json

    """
    H = empty_hypergraph()
    try:
        H._hypergraph.update(data["hypergraph-data"])
    except KeyError:
        raise XGIError("Failed to get hypergraph data attributes.")

    try:
        for id, dd in data["node-data"].items():
            if nodetype is not None:
                try:
                    id = nodetype(id)
                except ValueError as e:
                    raise TypeError(
                        f"Failed to convert edge IDs to type {nodetype}."
                    ) from e
            H.add_node(id, **dd)
    except KeyError:
        raise XGIError("Failed to import node attributes.")

    try:
        for id, edge in data["edge-dict"].items():
            if max_order and len(edge) > max_order + 1:
                continue
            if edgetype is not None:
                try:
                    id = edgetype(id)
                except ValueError as e:
                    raise TypeError(
                        f"Failed to convert the edge with ID {id} to type {edgetype}."
                    ) from e

            if nodetype is not None:
                try:
                    edge = {nodetype(n) for n in edge}
                except ValueError as e:
                    raise TypeError(
                        f"Failed to convert nodes to type {nodetype}."
                    ) from e
            H.add_edge(edge, id)

    except KeyError as e:
        raise XGIError("Failed to import edge dictionary.") from e

    try:
        if edgetype is None:
            edge_data = {
                key: val for key, val in data["edge-data"].items() if key in H.edges
            }
        else:
            edge_data = {
                edgetype(e): dd
                for e, dd in data["edge-data"].items()
                if edgetype(e) in H.edges
            }

        H.set_edge_attributes(edge_data)
    except KeyError as e:
        raise XGIError("Failed to import edge attributes.") from e

    return H



"""Methods for converting to and from an incidence matrix."""

from scipy.sparse import coo_array

from ..exception import XGIError
from ..generators import empty_hypergraph
from ..linalg import incidence_matrix

__all__ = ["from_incidence_matrix", "to_incidence_matrix"]


def from_incidence_matrix(d, create_using=None, nodelabels=None, edgelabels=None):
    """Create a hypergraph from an incidence matrix

    Parameters
    ----------
    d : numpy array or a scipy sparse arrary
        The incidence matrix where rows specify nodes and columns specify edges.
    create_using : Hypergraph constructor, optional
        The hypergraph object to add the data to, by default None
    nodelabels : list or 1D numpy array, optional
        The ordered list of node IDs to map the indices
        of the incidence matrix to, by default None
    edgelabels : list or 1D numpy array, optional
        The ordered list of edge IDs to map the indices
        of the incidence matrix to, by default None

    Returns
    -------
    Hypergraph object
        The constructed hypergraph

    Raises
    ------
    XGIError
        Raises an error if the specified labels are the wrong dimensions

    See Also
    --------
    incidence_matrix
    to_incidence_matrix
    """
    I = coo_array(d)
    n, m = I.shape

    if nodelabels is None:
        nodedict = dict(zip(range(n), range(n)))
    elif nodelabels is not None and len(nodelabels) != n:
        raise XGIError("Node dictionary is the wrong size.")
    else:
        nodedict = dict(zip(range(n), nodelabels))

    if edgelabels is None:
        edgedict = dict(zip(range(m), range(m)))
    elif edgelabels is not None and len(edgelabels) != m:
        raise XGIError("Edge dictionary is the wrong size.")
    else:
        edgedict = dict(zip(range(m), edgelabels))

    H = empty_hypergraph(create_using)

    for node, edge in zip(I.row, I.col):
        node = nodedict[node]
        edge = edgedict[edge]
        H.add_node_to_edge(edge, node)

    return H


def to_incidence_matrix(H, sparse=True, index=False):
    """Convert a hypergraph to an incidence matrix.

    Parameters
    ----------
    H : Hypergraph object
        The hypergraph of interest
    sparse : bool, optional
        Whether the constructed incidence matrix
        should be sparse, by default True
    index : bool, optional
        Whether to return the corresponding
        node and edge labels, by default False

    Returns
    -------
    numpy.ndarray or scipy csr_array
        The incidence matrix
    dict
        The dictionary mapping indices to node IDs, if index is True
    dict
        The dictionary mapping indices to edge IDs, if index is True

    See Also
    --------
    incidence_matrix
    from_incidence_matrix
    """
    return incidence_matrix(H, sparse=sparse, index=index)



"""Generators for some simple hypergraphs.

All the functions in this module return a Hypergraph class (i.e. a simple, undirected
hypergraph).

"""
from itertools import combinations

from ..exception import XGIError
from .classic import empty_hypergraph

__all__ = [
    "star_clique",
    "sunflower",
]


def star_clique(n_star, n_clique, d_max):
    """Generate a star-clique structure

    That is a star network and a clique network,
    connected by one pairwise edge connecting the centre of the star to the clique.
    network, the each clique is promoted to a hyperedge
    up to order d_max.

    Parameters
    ----------
    n_star : int
        Number of legs of the star
    n_clique : int
        Number of nodes in the clique
    d_max : int
        Maximum order up to which to promote
        cliques to hyperedges

    Returns
    -------
    H : Hypergraph

    Examples
    --------
    >>> import xgi
    >>> H = xgi.star_clique(6, 7, 2)

    Notes
    -----
    The total number of nodes is n_star + n_clique.

    """

    if n_star <= 0:
        raise ValueError("n_star must be an integer > 0.")
    if n_clique <= 0:
        raise ValueError("n_clique must be an integer > 0.")
    if d_max < 0:
        raise ValueError("d_max must be an integer >= 0.")
    elif d_max > n_clique - 1:
        raise ValueError("d_max must be <= n_clique - 1.")

    nodes_star = range(n_star)
    nodes_clique = range(n_star, n_star + n_clique)
    nodes = list(nodes_star) + list(nodes_clique)

    H = empty_hypergraph()
    H.add_nodes_from(nodes)

    # add star edges (center of the star is 0-th node)
    H.add_edges_from([(nodes_star[0], nodes_star[i]) for i in range(1, n_star)])

    # connect clique and star by adding last star leg
    H.add_edge((nodes_star[0], nodes_clique[0]))

    # add clique hyperedges up to order d_max
    H.add_edges_from(
        [e for d in range(1, d_max + 1) for e in combinations(nodes_clique, d + 1)]
    )

    return H


def sunflower(l, c, m):
    """Create a sunflower hypergraph.

    This creates an m-uniform hypergraph
    according to the sunflower model.

    Parameters
    ----------
    l : int
        Number of petals
    c : int
        Size of the core
    m : int
        Size of each edge

    Raises
    ------
    XGIError
        If the edge size is smaller than the core.

    Returns
    -------

    """
    from ..core import Hypergraph

    if m < c:
        raise XGIError("m cannot be smaller than c.")

    core_nodes = list(range(c))

    H = Hypergraph()
    start_label = c
    while start_label + (m - c) <= c + (m - c) * l:
        H.add_edge(core_nodes + [start_label + i for i in range(m - c)])
        start_label = start_label + (m - c)

    return H
